import numpy as np
import pytest

import autoarray as aa
from autoarray import exc


@pytest.fixture(name="simple_mask_7x7")
def make_simple_mask_7x7():

    mask = [
        [True, True, True, True, True, True, True],
        [True, True, True, True, True, True, True],
        [True, True, False, False, False, True, True],
        [True, True, False, False, False, True, True],
        [True, True, False, False, False, True, True],
        [True, True, True, True, True, True, True],
        [True, True, True, True, True, True, True],
    ]

    return aa.Mask2D.manual(mask=mask, pixel_scales=1.0, sub_size=1)


@pytest.fixture(name="simple_mask_5x5")
def make_simple_mask_5x5():

    mask = [
        [True, True, True, True, True],
        [True, False, False, False, True],
        [True, False, False, False, True],
        [True, False, False, False, True],
        [True, True, True, True, True],
    ]

    return aa.Mask2D.manual(mask=mask, pixel_scales=1.0, sub_size=1)


@pytest.fixture(name="simple_mask_index_array")
def make_simple_mask_index_array():
    return np.array([[6, 7, 8], [11, 12, 13], [16, 17, 18]])


@pytest.fixture(name="cross_mask")
def make_cross_mask():
    mask = np.full((5, 5), True)

    mask[2, 2] = False
    mask[1, 2] = False
    mask[3, 2] = False
    mask[2, 1] = False
    mask[2, 3] = False

    return aa.Mask2D.manual(mask=mask, pixel_scales=1.0, sub_size=1)


@pytest.fixture(name="cross_mask_index_array")
def make_cross_mask_index_array():
    return np.array([[-1, 0, -1], [1, 2, 3], [-1, 4, -1]])


@pytest.fixture(name="simple_image_frame_indexes")
def make_simple_image_frame_indexes(simple_convolver):
    return simple_convolver.make_image_frame_indexes((3, 3))


@pytest.fixture(name="cross_image_frame_indexes")
def make_cross_image_frame_indexes(cross_convolver):
    return cross_convolver.make_image_frame_indexes((3, 3))


@pytest.fixture(name="cross_mask_image_frame_indexes")
def make_cross_mask_image_frame_indexes(cross_convolver):
    return cross_convolver.make_blurring_image_frame_indexes((3, 3))


@pytest.fixture(name="simple_convolver")
def make_simple_convolver(simple_mask_5x5):

    return aa.Convolver(
        mask=simple_mask_5x5,
        kernel=aa.Kernel2D.manual_native(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], pixel_scales=1.0
        ),
    )


@pytest.fixture(name="cross_convolver")
def make_cross_convolver(cross_mask):
    return aa.Convolver(
        mask=cross_mask,
        kernel=aa.Kernel2D.manual_native(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], pixel_scales=1.0
        ),
    )


@pytest.fixture(name="simple_kernel")
def make_simple_kernel():
    return aa.Kernel2D.manual_native(
        kernel=[[0, 0.1, 0], [0.1, 0.6, 0.1], [0, 0.1, 0]], pixel_scales=1.0
    )


class TestNumbering:
    def test_simple_numbering(self, simple_mask_5x5, simple_mask_index_array):

        convolver = aa.Convolver(
            mask=simple_mask_5x5,
            kernel=aa.Kernel2D.ones(shape_native=(1, 1), pixel_scales=1.0),
        )

        mask_index_array = convolver.mask_index_array

        assert mask_index_array.shape == (5, 5)
        # noinspection PyUnresolvedReferences
        assert (
            mask_index_array
            == np.array(
                [
                    [-1, -1, -1, -1, -1],
                    [-1, 0, 1, 2, -1],
                    [-1, 3, 4, 5, -1],
                    [-1, 6, 7, 8, -1],
                    [-1, -1, -1, -1, -1],
                ]
            )
        ).all()

    def test__cross_mask(self, cross_mask):
        convolver = aa.Convolver(
            mask=cross_mask,
            kernel=aa.Kernel2D.ones(shape_native=(1, 1), pixel_scales=1.0),
        )

        assert (
            convolver.mask_index_array
            == np.array(
                [
                    [-1, -1, -1, -1, -1],
                    [-1, -1, 0, -1, -1],
                    [-1, 1, 2, 3, -1],
                    [-1, -1, 4, -1, -1],
                    [-1, -1, -1, -1, -1],
                ]
            )
        ).all()

    def test__even_kernel_failure(self):
        with pytest.raises(exc.ConvolverException):
            aa.Convolver(
                mask=np.full((3, 3), False),
                kernel=aa.Kernel2D.ones(shape_native=(2, 2), pixel_scales=1.0),
            )


class TestFrameExtraction:
    def test__frame_at_coords(self, simple_mask_5x5, simple_convolver):
        frame, kernel_frame = simple_convolver.frame_at_coordinates_jit(
            coordinates=(2, 2),
            mask=simple_mask_5x5,
            mask_index_array=simple_convolver.mask_index_array,
            kernel_2d=simple_convolver.kernel.native,
        )

        assert (frame == np.array([i for i in range(9)])).all()

        corner_frame = np.array([0, 1, 3, 4, -1, -1, -1, -1, -1])

        frame, kernel_frame = simple_convolver.frame_at_coordinates_jit(
            coordinates=(1, 1),
            mask=simple_mask_5x5,
            mask_index_array=simple_convolver.mask_index_array,
            kernel_2d=simple_convolver.kernel.native,
        )

        assert (frame == corner_frame).all()

    def test__kernel_frame_at_coords(self, simple_mask_5x5, simple_convolver):

        frame, kernel_frame = simple_convolver.frame_at_coordinates_jit(
            coordinates=(2, 2),
            mask=simple_mask_5x5,
            mask_index_array=simple_convolver.mask_index_array,
            kernel_2d=simple_convolver.kernel.native,
        )

        assert (
            kernel_frame == np.array([[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])
        ).all()

        frame, kernel_frame = simple_convolver.frame_at_coordinates_jit(
            coordinates=(1, 1),
            mask=simple_mask_5x5,
            mask_index_array=simple_convolver.mask_index_array,
            kernel_2d=simple_convolver.kernel.native,
        )

        assert (
            kernel_frame == np.array([5.0, 6.0, 8.0, 9.0, -1, -1, -1, -1, -1])
        ).all()

    def test__simple_square(self, simple_convolver):
        assert 9 == len(simple_convolver.image_frame_1d_indexes)

        assert (
            simple_convolver.image_frame_1d_indexes[4]
            == np.array([i for i in range(9)])
        ).all()

    def test__frame_5x5_kernel__at_coords(self, simple_mask_7x7):
        convolver = aa.Convolver(
            mask=simple_mask_7x7,
            kernel=aa.Kernel2D.manual_native(
                [
                    [1.0, 2.0, 3.0, 4.0, 5.0],
                    [6.0, 7.0, 8.0, 9.0, 10.0],
                    [11.0, 12.0, 13.0, 14.0, 15.0],
                    [16.0, 17.0, 18.0, 19.0, 20.0],
                    [21.0, 22.0, 23.0, 24.0, 25.0],
                ],
                pixel_scales=1.0,
            ),
        )

        frame, kernel_frame = convolver.frame_at_coordinates_jit(
            coordinates=(2, 2),
            mask=simple_mask_7x7,
            mask_index_array=convolver.mask_index_array,
            kernel_2d=convolver.kernel.native,
        )

        assert (
            kernel_frame
            == np.array(
                [
                    13.0,
                    14.0,
                    15.0,
                    18.0,
                    19.0,
                    20.0,
                    23.0,
                    24.0,
                    25.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                ]
            )
        ).all()

        frame, kernel_frame = convolver.frame_at_coordinates_jit(
            coordinates=(3, 2),
            mask=simple_mask_7x7,
            mask_index_array=convolver.mask_index_array,
            kernel_2d=convolver.kernel.native,
        )

        assert (
            kernel_frame
            == np.array(
                [
                    8.0,
                    9.0,
                    10.0,
                    13.0,
                    14.0,
                    15.0,
                    18.0,
                    19.0,
                    20.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1,
                ]
            )
        ).all()

        frame, kernel_frame = convolver.frame_at_coordinates_jit(
            coordinates=(3, 3),
            mask=simple_mask_7x7,
            mask_index_array=convolver.mask_index_array,
            kernel_2d=convolver.kernel.native,
        )

        assert (
            kernel_frame
            == np.array(
                [
                    7.0,
                    8.0,
                    9.0,
                    12.0,
                    13.0,
                    14.0,
                    17.0,
                    18.0,
                    19.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1.0,
                    -1,
                ]
            )
        ).all()


class TestImageFrameIndexes:
    def test__masked_cross__3x3_kernel(self, cross_convolver):
        assert 5 == len(cross_convolver.image_frame_1d_indexes)

        assert (
            cross_convolver.image_frame_1d_indexes[0]
            == np.array([0, 1, 2, 3, -1, -1, -1, -1, -1])
        ).all()
        assert (
            cross_convolver.image_frame_1d_indexes[1]
            == np.array([0, 1, 2, 4, -1, -1, -1, -1, -1])
        ).all()
        assert (
            cross_convolver.image_frame_1d_indexes[2]
            == np.array([0, 1, 2, 3, 4, -1, -1, -1, -1])
        ).all()
        assert (
            cross_convolver.image_frame_1d_indexes[3]
            == np.array([0, 2, 3, 4, -1, -1, -1, -1, -1])
        ).all()
        assert (
            cross_convolver.image_frame_1d_indexes[4]
            == np.array([1, 2, 3, 4, -1, -1, -1, -1, -1])
        ).all()

    def test__masked_square__3x5_kernel__loses_edge_of_top_and_bottom_rows(
        self, simple_mask_7x7
    ):
        convolver = aa.Convolver(
            mask=simple_mask_7x7,
            kernel=aa.Kernel2D.ones(shape_native=(3, 5), pixel_scales=1.0),
        )

        assert (
            convolver.image_frame_1d_indexes[0]
            == np.array([0, 1, 2, 3, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[1]
            == np.array([0, 1, 2, 3, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[2]
            == np.array([0, 1, 2, 3, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[3]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[4]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[5]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[6]
            == np.array([3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[7]
            == np.array([3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[8]
            == np.array([3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()

    def test__masked_square__5x3_kernel__loses_edge_of_left_and_right_columns(
        self, simple_mask_7x7
    ):
        convolver = aa.Convolver(
            mask=simple_mask_7x7,
            kernel=aa.Kernel2D.ones(shape_native=(5, 3), pixel_scales=1.0),
        )

        assert (
            convolver.image_frame_1d_indexes[0]
            == np.array([0, 1, 3, 4, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[1]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[2]
            == np.array([1, 2, 4, 5, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[3]
            == np.array([0, 1, 3, 4, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[4]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[5]
            == np.array([1, 2, 4, 5, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[6]
            == np.array([0, 1, 3, 4, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[7]
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.image_frame_1d_indexes[8]
            == np.array([1, 2, 4, 5, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1])
        ).all()

    def test__masked_square__7x7_kernel(self, simple_mask_7x7):
        convolver = aa.Convolver(
            mask=simple_mask_7x7,
            kernel=aa.Kernel2D.ones(shape_native=(5, 5), pixel_scales=1.0),
        )

        assert (
            convolver.image_frame_1d_indexes[0]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[1]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[2]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[3]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[4]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[5]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[6]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[7]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_indexes[8]
            == np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()


class TestImageFrameKernels:
    def test_simple_square(self, simple_convolver):
        assert 9 == len(simple_convolver.image_frame_1d_indexes)

        assert (
            simple_convolver.image_frame_1d_kernels[4]
            == np.array([[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])
        ).all()

    def test_masked_square__3x5_kernel__loses_edge_of_top_and_bottom_rows(
        self, simple_mask_7x7
    ):
        convolver = aa.Convolver(
            mask=simple_mask_7x7,
            kernel=aa.Kernel2D.manual_native(
                [
                    [1.0, 2.0, 3.0, 4.0, 5.0],
                    [6.0, 7.0, 8.0, 9.0, 10.0],
                    [11.0, 12.0, 13.0, 14.0, 15.0],
                ],
                pixel_scales=1.0,
            ),
        )

        assert (
            convolver.image_frame_1d_kernels[0]
            == np.array(
                [8.0, 9.0, 10.0, 13.0, 14.0, 15.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[1]
            == np.array(
                [7.0, 8.0, 9.0, 12.0, 13.0, 14.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[2]
            == np.array(
                [6.0, 7.0, 8.0, 11.0, 12.0, 13.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[3]
            == np.array(
                [
                    3.0,
                    4.0,
                    5.0,
                    8.0,
                    9.0,
                    10.0,
                    13.0,
                    14.0,
                    15.0,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[4]
            == np.array(
                [2.0, 3.0, 4.0, 7.0, 8.0, 9.0, 12.0, 13.0, 14.0, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[5]
            == np.array(
                [1.0, 2.0, 3.0, 6.0, 7.0, 8.0, 11.0, 12.0, 13.0, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[6]
            == np.array(
                [3.0, 4.0, 5.0, 8.0, 9.0, 10.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[7]
            == np.array(
                [2.0, 3.0, 4.0, 7.0, 8.0, 9.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[8]
            == np.array(
                [1.0, 2.0, 3.0, 6.0, 7.0, 8.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()

    def test_masked_square__5x3_kernel__loses_edge_of_left_and_right_columns(
        self, simple_mask_7x7
    ):
        convolver = aa.Convolver(
            mask=simple_mask_7x7,
            kernel=aa.Kernel2D.manual_native(
                [
                    [1.0, 2.0, 3.0],
                    [4.0, 5.0, 6.0],
                    [7.0, 8.0, 9.0],
                    [10.0, 11.0, 12.0],
                    [13.0, 14.0, 15.0],
                ],
                pixel_scales=1.0,
            ),
        )

        assert (
            convolver.image_frame_1d_kernels[0]
            == np.array(
                [8.0, 9.0, 11.0, 12.0, 14.0, 15.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[1]
            == np.array(
                [
                    7.0,
                    8.0,
                    9.0,
                    10.0,
                    11.0,
                    12.0,
                    13.0,
                    14.0,
                    15.0,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[2]
            == np.array(
                [7.0, 8.0, 10.0, 11.0, 13.0, 14.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[3]
            == np.array(
                [5.0, 6.0, 8.0, 9.0, 11.0, 12.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[4]
            == np.array(
                [4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[5]
            == np.array(
                [4.0, 5.0, 7.0, 8.0, 10.0, 11.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[6]
            == np.array(
                [2.0, 3.0, 5.0, 6.0, 8.0, 9.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[7]
            == np.array(
                [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, -1, -1, -1, -1, -1, -1]
            )
        ).all()
        assert (
            convolver.image_frame_1d_kernels[8]
            == np.array(
                [1.0, 2.0, 4.0, 5.0, 7.0, 8.0, -1, -1, -1, -1, -1, -1, -1, -1, -1]
            )
        ).all()


class TestBlurringFrameIndxes:
    def test__blurring_region_3x3_kernel(self, cross_mask):

        convolver = aa.Convolver(
            mask=cross_mask,
            kernel=aa.Kernel2D.ones(shape_native=(3, 3), pixel_scales=1.0),
        )

        assert (
            convolver.blurring_frame_1d_indexes[4]
            == np.array([0, 1, 2, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[5]
            == np.array([0, 2, 3, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[10]
            == np.array([1, 2, 4, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_indexes[11]
            == np.array([2, 3, 4, -1, -1, -1, -1, -1, -1])
        ).all()


class TestBlurringFrameKernels:
    def test__blurring_region_3x3_kernel(self, cross_mask):

        convolver = aa.Convolver(
            mask=cross_mask,
            kernel=aa.Kernel2D.manual_native(
                [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], pixel_scales=1.0
            ),
        )

        assert (
            convolver.blurring_frame_1d_kernels[4]
            == np.array([6.0, 8.0, 9.0, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_kernels[5]
            == np.array([4.0, 7.0, 8.0, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_kernels[10]
            == np.array([2.0, 3.0, 6.0, -1, -1, -1, -1, -1, -1])
        ).all()
        assert (
            convolver.blurring_frame_1d_kernels[11]
            == np.array([1.0, 2.0, 4.0, -1, -1, -1, -1, -1, -1])
        ).all()


class TestFrameLengths:
    def test__frames_are_from_examples_above__lengths_are_right(self, simple_mask_7x7):
        convolver = aa.Convolver(
            mask=simple_mask_7x7,
            kernel=aa.Kernel2D.ones(shape_native=(3, 5), pixel_scales=1.0),
        )

        # convolver_image.image_frame_indexes[0] == np.array([0, 1, 2, 3, 4, 5])
        # convolver_image.image_frame_indexes[1] == np.array([0, 1, 2, 3, 4, 5])
        # convolver_image.image_frame_indexes[2] == np.array([0, 1, 2, 3, 4, 5])
        # convolver_image.image_frame_indexes[3] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        # (convolver_image.image_frame_indexes[4] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[5] == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[6] == np.array([3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[7] == np.array([3, 4, 5, 6, 7, 8])
        # convolver_image.image_frame_indexes[8] == np.array([3, 4, 5, 6, 7, 8])

        assert (
            convolver.image_frame_1d_lengths == np.array([6, 6, 6, 9, 9, 9, 6, 6, 6])
        ).all()


class TestConvolveMappingMatrix:
    def test__asymetric_convolver__matrix_blurred_correctly(self):

        mask = np.array(
            [
                [True, True, True, True, True, True],
                [True, False, False, False, False, True],
                [True, False, False, False, False, True],
                [True, False, False, False, False, True],
                [True, False, False, False, False, True],
                [True, True, True, True, True, True],
            ]
        )

        asymmetric_kernel = aa.Kernel2D.manual_native(
            array=[[0, 0.0, 0], [0.4, 0.2, 0.3], [0, 0.1, 0]], pixel_scales=1.0
        )

        convolver = aa.Convolver(mask=mask, kernel=asymmetric_kernel)

        mapping = np.array(
            [
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [
                    0,
                    1,
                    0,
                ],  # The 0.3 should be 'chopped' from this pixel as it is on the right-most edge
                [0, 0, 0],
                [1, 0, 0],
                [0, 0, 1],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
            ]
        )

        blurred_mapping = convolver.convolve_mapping_matrix(mapping)

        assert (
            blurred_mapping
            == np.array(
                [
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0.4, 0],
                    [0, 0.2, 0],
                    [0.4, 0, 0],
                    [0.2, 0, 0.4],
                    [0.3, 0, 0.2],
                    [0, 0.1, 0.3],
                    [0, 0, 0],
                    [0.1, 0, 0],
                    [0, 0, 0.1],
                    [0, 0, 0],
                ]
            )
        ).all()

    def test__asymetric_convolver__multiple_overlapping_blurred_entires_in_matrix(self):

        mask = np.array(
            [
                [True, True, True, True, True, True],
                [True, False, False, False, False, True],
                [True, False, False, False, False, True],
                [True, False, False, False, False, True],
                [True, False, False, False, False, True],
                [True, True, True, True, True, True],
            ]
        )

        asymmetric_kernel = aa.Kernel2D.manual_native(
            array=[[0, 0.0, 0], [0.4, 0.2, 0.3], [0, 0.1, 0]], pixel_scales=1.0
        )

        convolver = aa.Convolver(mask=mask, kernel=asymmetric_kernel)

        mapping = np.array(
            [
                [0, 1, 0],
                [0, 1, 0],
                [0, 1, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [
                    0,
                    1,
                    0,
                ],  # The 0.3 should be 'chopped' from this pixel as it is on the right-most edge
                [1, 0, 0],
                [1, 0, 0],
                [0, 0, 1],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
            ]
        )

        blurred_mapping = convolver.convolve_mapping_matrix(mapping)

        assert blurred_mapping == pytest.approx(
            np.array(
                [
                    [0, 0.6, 0],
                    [0, 0.9, 0],
                    [0, 0.5, 0],
                    [0, 0.3, 0],
                    [0, 0.1, 0],
                    [0, 0.1, 0],
                    [0, 0.5, 0],
                    [0, 0.2, 0],
                    [0.6, 0, 0],
                    [0.5, 0, 0.4],
                    [0.3, 0, 0.2],
                    [0, 0.1, 0.3],
                    [0.1, 0, 0],
                    [0.1, 0, 0],
                    [0, 0, 0.1],
                    [0, 0, 0],
                ]
            ),
            1e-4,
        )


class TestConvolution:
    def test__cross_mask_with_blurring_entries__returns_array(self):

        cross_mask = aa.Mask2D.manual(
            mask=[
                [True, True, True, True, True],
                [True, True, False, True, True],
                [True, False, False, False, True],
                [True, True, False, True, True],
                [True, True, True, True, True],
            ],
            pixel_scales=0.1,
            sub_size=1,
        )

        kernel = aa.Kernel2D.manual_native(
            array=[[0, 0.2, 0], [0.2, 0.4, 0.2], [0, 0.2, 0]], pixel_scales=0.1
        )

        convolver = aa.Convolver(mask=cross_mask, kernel=kernel)

        image_array = aa.Array2D.manual_mask(array=[1, 0, 0, 0, 0], mask=cross_mask)

        blurring_mask = cross_mask.blurring_mask_from_kernel_shape(
            kernel_shape_native=kernel.shape_native
        )

        blurring_array = aa.Array2D.manual_mask(
            array=[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], mask=blurring_mask
        )

        result = convolver.convolved_image_from_image_and_blurring_image(
            image=image_array, blurring_image=blurring_array
        )

        assert (np.round(result, 1) == np.array([0.6, 0.2, 0.2, 0.0, 0.0])).all()


class TestCompareToFull2dConv:
    def test__compare_convolver_to_2d_convolution(self):
        # Setup a blurred datas_, using the PSF to perform the convolution in 2D, then masks it to make a 1d array.

        import scipy.signal

        kernel = aa.Kernel2D.manual_native(
            array=np.arange(49).reshape(7, 7), pixel_scales=1.0
        )
        image = aa.Array2D.manual_native(
            array=np.arange(900).reshape(30, 30), pixel_scales=1.0
        )
        blurred_image = scipy.signal.convolve2d(
            image.native, kernel.native, mode="same"
        )
        blurred_image = aa.Array2D.manual_native(array=blurred_image, pixel_scales=1.0)

        mask = aa.Mask2D.circular(
            shape_native=(30, 30), pixel_scales=(1.0, 1.0), sub_size=1, radius=4.0
        )

        masked_image = aa.Array2D.manual_mask(
            array=image.native, mask=mask, store_slim=True
        )

        blurred_masked_image = aa.Array2D.manual_mask(
            array=blurred_image.native, mask=mask, store_slim=True
        )

        # Now reproduce this datas_ using the frame convolver_image

        blurring_mask = mask.blurring_mask_from_kernel_shape(
            kernel_shape_native=kernel.shape_native
        )
        convolver = aa.Convolver(mask=mask, kernel=kernel)

        blurring_image = aa.Array2D.manual_mask(
            array=image.native, mask=blurring_mask, store_slim=True
        )

        blurred_masked_im_1 = convolver.convolved_image_from_image_and_blurring_image(
            image=masked_image, blurring_image=blurring_image
        )

        assert blurred_masked_image == pytest.approx(blurred_masked_im_1, 1e-4)



from . import array_util as array
from . import binning_util as binning
from . import fit_util as fit
from . import frame_util as frame
from . import geometry_util as geometry
from . import grid_util as grid
from . import inversion_util as inversion
from . import mapper_util as mapper
from . import mask_util as mask
from . import pixelization_util as pixelization
from . import regularization_util as regularization
from . import sparse_util as sparse
from . import transformer_util as transformer



import os
from os import path

import numpy as np
import pytest
import shutil

import autoarray as aa
from autoarray import exc

test_data_dir = path.join(
    "{}".format(path.dirname(path.realpath(__file__))), "files", "array"
)


class TestConstructorMethods:
    def test__constructor_class_method_native__store_slim(self):

        arr = aa.Array2D.manual_native(
            array=np.ones((3, 3)), sub_size=1, pixel_scales=(1.0, 1.0), store_slim=True
        )

        assert (arr == np.ones((9,))).all()
        assert (arr.slim == np.ones((9,))).all()
        assert (arr.native == np.ones((3, 3))).all()
        assert (arr.slim_binned == np.ones((9,))).all()
        assert (arr.native_binned == np.ones((3, 3))).all()
        assert (arr.binned == np.ones((9,))).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.mask.central_pixel_coordinates == (1.0, 1.0)
        assert arr.mask.shape_native_scaled == pytest.approx((3.0, 3.0))
        assert arr.mask.scaled_maxima == (1.5, 1.5)
        assert arr.mask.scaled_minima == (-1.5, -1.5)

        arr = aa.Array2D.manual_native(
            array=np.ones((4, 4)), sub_size=2, pixel_scales=(0.1, 0.1), store_slim=True
        )

        assert (arr == np.ones((16,))).all()
        assert (arr.slim == np.ones((16,))).all()
        assert (arr.native == np.ones((4, 4))).all()
        assert (arr.slim_binned == np.ones((4,))).all()
        assert (arr.native_binned == np.ones((2, 2))).all()
        assert (arr.binned == np.ones((4,))).all()
        assert arr.pixel_scales == (0.1, 0.1)
        assert arr.mask.central_pixel_coordinates == (0.5, 0.5)
        assert arr.mask.shape_native_scaled == pytest.approx((0.2, 0.2))
        assert arr.mask.scaled_maxima == pytest.approx((0.1, 0.1), 1e-4)
        assert arr.mask.scaled_minima == pytest.approx((-0.1, -0.1), 1e-4)

        arr = aa.Array2D.manual_native(
            array=np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]]),
            pixel_scales=(0.1, 0.1),
            sub_size=2,
            origin=(1.0, 1.0),
            store_slim=True,
        )

        assert (arr == np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])).all()
        assert arr.shape_native == (2, 1)
        assert arr.sub_shape_native == (4, 2)
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])).all()
        assert (
            arr.native == np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])
        ).all()
        assert arr.native_binned.shape == (2, 1)
        assert (arr.slim_binned == np.array([2.5, 6.5])).all()
        assert (arr.native_binned == np.array([[2.5], [6.5]])).all()
        assert (arr.binned == np.array([2.5, 6.5])).all()
        assert arr.pixel_scales == (0.1, 0.1)
        assert arr.mask.central_pixel_coordinates == (0.5, 0.0)
        assert arr.mask.shape_native_scaled == pytest.approx((0.2, 0.1))
        assert arr.mask.scaled_maxima == pytest.approx((1.1, 1.05), 1e-4)
        assert arr.mask.scaled_minima == pytest.approx((0.9, 0.95), 1e-4)

        arr = aa.Array2D.manual_native(
            array=np.ones((3, 3)),
            pixel_scales=(2.0, 1.0),
            sub_size=1,
            origin=(-1.0, -2.0),
            store_slim=True,
        )

        assert arr == pytest.approx(np.ones((9,)), 1e-4)
        assert arr.slim == pytest.approx(np.ones((9,)), 1e-4)
        assert arr.native == pytest.approx(np.ones((3, 3)), 1e-4)
        assert arr.pixel_scales == (2.0, 1.0)
        assert arr.mask.central_pixel_coordinates == (1.0, 1.0)
        assert arr.mask.shape_native_scaled == pytest.approx((6.0, 3.0))
        assert arr.origin == (-1.0, -2.0)
        assert arr.mask.scaled_maxima == pytest.approx((2.0, -0.5), 1e-4)
        assert arr.mask.scaled_minima == pytest.approx((-4.0, -3.5), 1e-4)

    def test__constructor_class_method_slim__store_slim(self):
        arr = aa.Array2D.manual_slim(
            array=np.ones((9,)),
            shape_native=(3, 3),
            pixel_scales=(2.0, 1.0),
            sub_size=1,
            origin=(-1.0, -2.0),
            store_slim=True,
        )

        assert arr == pytest.approx(np.ones((9,)), 1e-4)
        assert arr.slim == pytest.approx(np.ones((9,)), 1e-4)
        assert arr.native == pytest.approx(np.ones((3, 3)), 1e-4)
        assert arr.pixel_scales == (2.0, 1.0)
        assert arr.mask.central_pixel_coordinates == (1.0, 1.0)
        assert arr.mask.shape_native_scaled == pytest.approx((6.0, 3.0))
        assert arr.origin == (-1.0, -2.0)
        assert arr.mask.scaled_maxima == pytest.approx((2.0, -0.5), 1e-4)
        assert arr.mask.scaled_minima == pytest.approx((-4.0, -3.5), 1e-4)

    def test__constructor_class_method_native__store_native(self):

        arr = aa.Array2D.manual_native(
            array=np.ones((3, 3)), sub_size=1, pixel_scales=(1.0, 1.0), store_slim=False
        )

        assert (arr == np.ones((3, 3))).all()
        assert (arr.slim == np.ones((9,))).all()
        assert (arr.native == np.ones((3, 3))).all()
        assert (arr.slim_binned == np.ones((9,))).all()
        assert (arr.native_binned == np.ones((3, 3))).all()
        assert (arr.binned == np.ones((3, 3))).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.mask.central_pixel_coordinates == (1.0, 1.0)
        assert arr.mask.shape_native_scaled == pytest.approx((3.0, 3.0))
        assert arr.mask.scaled_maxima == (1.5, 1.5)
        assert arr.mask.scaled_minima == (-1.5, -1.5)

        arr = aa.Array2D.manual_native(
            array=np.ones((4, 4)), sub_size=2, pixel_scales=(0.1, 0.1), store_slim=False
        )

        assert (arr == np.ones((4, 4))).all()
        assert (arr.slim == np.ones((16,))).all()
        assert (arr.native == np.ones((4, 4))).all()
        assert (arr.slim_binned == np.ones((4,))).all()
        assert (arr.native_binned == np.ones((2, 2))).all()
        assert (arr.binned == np.ones((2, 2))).all()
        assert arr.pixel_scales == (0.1, 0.1)
        assert arr.mask.central_pixel_coordinates == (0.5, 0.5)
        assert arr.mask.shape_native_scaled == pytest.approx((0.2, 0.2))
        assert arr.mask.scaled_maxima == pytest.approx((0.1, 0.1), 1e-4)
        assert arr.mask.scaled_minima == pytest.approx((-0.1, -0.1), 1e-4)

        arr = aa.Array2D.manual_native(
            array=np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]]),
            pixel_scales=(0.1, 0.1),
            sub_size=2,
            origin=(1.0, 1.0),
            store_slim=False,
        )

        assert (arr == np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])).all()
        assert arr.shape_native == (2, 1)
        assert arr.sub_shape_native == (4, 2)
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])).all()
        assert (
            arr.native == np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])
        ).all()
        assert arr.native_binned.shape == (2, 1)
        assert (arr.slim_binned == np.array([2.5, 6.5])).all()
        assert (arr.native_binned == np.array([[2.5], [6.5]])).all()
        assert (arr.binned == np.array([[2.5], [6.5]])).all()
        assert arr.pixel_scales == (0.1, 0.1)
        assert arr.mask.central_pixel_coordinates == (0.5, 0.0)
        assert arr.mask.shape_native_scaled == pytest.approx((0.2, 0.1))
        assert arr.mask.scaled_maxima == pytest.approx((1.1, 1.05), 1e-4)
        assert arr.mask.scaled_minima == pytest.approx((0.9, 0.95), 1e-4)

        arr = aa.Array2D.manual_native(
            array=np.ones((3, 3)),
            pixel_scales=(2.0, 1.0),
            sub_size=1,
            origin=(-1.0, -2.0),
            store_slim=False,
        )

        assert arr == pytest.approx(np.ones((3, 3)), 1e-4)
        assert arr.slim == pytest.approx(np.ones((9,)), 1e-4)
        assert arr.native == pytest.approx(np.ones((3, 3)), 1e-4)
        assert arr.pixel_scales == (2.0, 1.0)
        assert arr.mask.central_pixel_coordinates == (1.0, 1.0)
        assert arr.mask.shape_native_scaled == pytest.approx((6.0, 3.0))
        assert arr.origin == (-1.0, -2.0)
        assert arr.mask.scaled_maxima == pytest.approx((2.0, -0.5), 1e-4)
        assert arr.mask.scaled_minima == pytest.approx((-4.0, -3.5), 1e-4)

    def test__constructor_class_method_slim__store_native(self):
        arr = aa.Array2D.manual_slim(
            array=np.ones((9,)),
            shape_native=(3, 3),
            pixel_scales=(2.0, 1.0),
            sub_size=1,
            origin=(-1.0, -2.0),
            store_slim=False,
        )

        assert arr == pytest.approx(np.ones((3, 3)), 1e-4)
        assert arr.slim == pytest.approx(np.ones((9,)), 1e-4)
        assert arr.native == pytest.approx(np.ones((3, 3)), 1e-4)
        assert arr.pixel_scales == (2.0, 1.0)
        assert arr.mask.central_pixel_coordinates == (1.0, 1.0)
        assert arr.mask.shape_native_scaled == pytest.approx((6.0, 3.0))
        assert arr.origin == (-1.0, -2.0)
        assert arr.mask.scaled_maxima == pytest.approx((2.0, -0.5), 1e-4)
        assert arr.mask.scaled_minima == pytest.approx((-4.0, -3.5), 1e-4)


class TestNewArrays:
    def test__pad__compare_to_array_util(self):
        array_2d = np.ones((5, 5))
        array_2d[2, 2] = 2.0

        arr = aa.Array2D.manual_native(
            array=array_2d, sub_size=1, pixel_scales=(1.0, 1.0)
        )

        arr = arr.resized_from(new_shape=(7, 7))

        arr_resized_manual = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 2.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == arr_resized_manual).all()
        assert arr.mask.pixel_scales == (1.0, 1.0)

    def test__trim__compare_to_array_util(self):
        array_2d = np.ones((5, 5))
        array_2d[2, 2] = 2.0

        arr = aa.Array2D.manual_native(
            array=array_2d, sub_size=1, pixel_scales=(1.0, 1.0)
        )

        arr = arr.resized_from(new_shape=(3, 3))

        arr_resized_manual = np.array(
            [[1.0, 1.0, 1.0], [1.0, 2.0, 1.0], [1.0, 1.0, 1.0]]
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == arr_resized_manual).all()
        assert arr.mask.pixel_scales == (1.0, 1.0)

    def test__padded_from_kernel_shape__padded_edge_of_zeros_where_extra_psf_blurring_is_performed(
        self,
    ):

        array_2d = np.ones((5, 5))
        array_2d[2, 2] = 2.0

        arr = aa.Array2D.manual_native(
            array=array_2d, sub_size=1, pixel_scales=(1.0, 1.0)
        )

        new_arr = arr.padded_before_convolution_from(kernel_shape=(3, 3))

        assert type(new_arr) == aa.Array2D
        assert new_arr.native[0, 0] == 0.0
        assert new_arr.shape_native == (7, 7)
        assert new_arr.mask.pixel_scales == (1.0, 1.0)

        new_arr = arr.padded_before_convolution_from(kernel_shape=(5, 5))

        assert type(new_arr) == aa.Array2D
        assert new_arr.native[0, 0] == 0.0
        assert new_arr.shape_native == (9, 9)
        assert new_arr.mask.pixel_scales == (1.0, 1.0)

        array_2d = np.ones((9, 9))
        array_2d[4, 4] = 2.0

        arr = aa.Array2D.manual_native(
            array=array_2d, sub_size=1, pixel_scales=(1.0, 1.0)
        )

        new_arr = arr.padded_before_convolution_from(kernel_shape=(7, 7))

        assert type(new_arr) == aa.Array2D
        assert new_arr.native[0, 0] == 0.0
        assert new_arr.shape_native == (15, 15)
        assert new_arr.mask.pixel_scales == (1.0, 1.0)

    def test__trimmed_from_kernel_shape__trim_edges_where_extra_psf_blurring_is_performed(
        self,
    ):
        array_2d = np.ones((5, 5))
        array_2d[2, 2] = 2.0

        arr = aa.Array2D.manual_native(
            array=array_2d, sub_size=1, pixel_scales=(1.0, 1.0)
        )

        new_arr = arr.trimmed_after_convolution_from(kernel_shape=(3, 3))

        assert type(new_arr) == aa.Array2D
        assert (
            new_arr.native
            == np.array([[1.0, 1.0, 1.0], [1.0, 2.0, 1.0], [1.0, 1.0, 1.0]])
        ).all()
        assert new_arr.mask.pixel_scales == (1.0, 1.0)

        new_arr = arr.trimmed_after_convolution_from(kernel_shape=(5, 5))

        assert type(new_arr) == aa.Array2D
        assert (new_arr.native == np.array([[2.0]])).all()
        assert new_arr.mask.pixel_scales == (1.0, 1.0)

        array_2d = np.ones((9, 9))
        array_2d[4, 4] = 2.0

        arr = aa.Array2D.manual_native(
            array=array_2d, sub_size=1, pixel_scales=(1.0, 1.0)
        )

        new_arr = arr.trimmed_after_convolution_from(kernel_shape=(7, 7))

        assert type(new_arr) == aa.Array2D
        assert (
            new_arr.native
            == np.array([[1.0, 1.0, 1.0], [1.0, 2.0, 1.0], [1.0, 1.0, 1.0]])
        ).all()
        assert new_arr.mask.pixel_scales == (1.0, 1.0)

    def test__zoomed__2d_array_zoomed__uses_the_limits_of_the_mask(self):

        array_2d = [
            [1.0, 2.0, 3.0, 4.0],
            [5.0, 6.0, 7.0, 8.0],
            [9.0, 10.0, 11.0, 12.0],
            [13.0, 14.0, 15.0, 16.0],
        ]

        mask = aa.Mask2D.manual(
            mask=[
                [True, True, True, True],
                [True, False, False, True],
                [True, False, False, True],
                [True, True, True, True],
            ],
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )

        arr_masked = aa.Array2D.manual_mask(array=array_2d, mask=mask)

        arr_zoomed = arr_masked.zoomed_around_mask(buffer=0)

        assert (arr_zoomed.native == np.array([[6.0, 7.0], [10.0, 11.0]])).all()

        mask = aa.Mask2D.manual(
            mask=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, False],
                    [True, False, False, False],
                    [True, True, True, True],
                ]
            ),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )

        arr_masked = aa.Array2D.manual_mask(array=array_2d, mask=mask)
        arr_zoomed = arr_masked.zoomed_around_mask(buffer=0)
        assert (
            arr_zoomed.native == np.array([[6.0, 7.0, 8.0], [10.0, 11.0, 12.0]])
        ).all()

        mask = aa.Mask2D.manual(
            mask=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, False, False, True],
                    [True, False, False, True],
                ]
            ),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )

        arr_masked = aa.Array2D.manual_mask(array=array_2d, mask=mask)
        arr_zoomed = arr_masked.zoomed_around_mask(buffer=0)
        assert (
            arr_zoomed.native == np.array([[6.0, 7.0], [10.0, 11.0], [14.0, 15.0]])
        ).all()

        mask = aa.Mask2D.manual(
            mask=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [False, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )

        arr_masked = aa.Array2D.manual_mask(array=array_2d, mask=mask)
        arr_zoomed = arr_masked.zoomed_around_mask(buffer=0)

        assert (
            arr_zoomed.native == np.array([[0.0, 6.0, 7.0], [9.0, 10.0, 11.0]])
        ).all()

        mask = aa.Mask2D.manual(
            mask=np.array(
                [
                    [True, False, True, True],
                    [True, False, False, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )

        arr_masked = aa.Array2D.manual_mask(array=array_2d, mask=mask)
        arr_zoomed = arr_masked.zoomed_around_mask(buffer=0)
        assert (
            arr_zoomed.native == np.array([[2.0, 0.0], [6.0, 7.0], [10.0, 11.0]])
        ).all()

        mask = aa.Mask2D.manual(
            mask=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )

        arr_masked = aa.Array2D.manual_mask(array=array_2d, mask=mask)
        arr_zoomed = arr_masked.zoomed_around_mask(buffer=1)

        assert (
            arr_zoomed.native
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 6.0, 7.0, 0.0],
                    [0.0, 10.0, 11.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

    def test__zoomed_2d_array_zoomed__centre_is_updated_using_original_mask(self):

        array_2d = np.ones(shape=(4, 4))

        mask = aa.Mask2D.manual(
            mask=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )

        arr_masked = aa.Array2D.manual_mask(array=array_2d, mask=mask)

        arr_zoomed = arr_masked.zoomed_around_mask(buffer=0)

        assert arr_zoomed.mask.origin == (0.0, 0.0)

        array_2d = np.ones(shape=(6, 6))

        mask = aa.Mask2D.manual(
            mask=np.array(
                [
                    [True, True, True, True, True, True],
                    [True, True, True, False, False, True],
                    [True, True, True, False, False, True],
                    [True, True, True, True, True, True],
                    [True, True, True, True, True, True],
                    [True, True, True, True, True, True],
                ]
            ),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )

        arr_masked = aa.Array2D.manual_mask(array=array_2d, mask=mask)

        arr_zoomed = arr_masked.zoomed_around_mask(buffer=0)

        assert arr_zoomed.mask.origin == (1.0, 1.0)

        mask = aa.Mask2D.manual(
            mask=np.array(
                [
                    [True, True, True, True, True, True],
                    [True, True, True, True, True, True],
                    [True, True, True, False, False, True],
                    [True, True, True, False, False, True],
                    [True, True, True, True, True, True],
                    [True, True, True, True, True, True],
                ]
            ),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )

        arr_masked = aa.Array2D.manual_mask(array=array_2d, mask=mask)

        arr_zoomed = arr_masked.zoomed_around_mask(buffer=0)

        assert arr_zoomed.mask.origin == (0.0, 1.0)

    def test__zoomed__array_extent__uses_the_limits_of_the_unzoomed_mask(self):

        array_2d = [
            [1.0, 2.0, 3.0, 4.0],
            [5.0, 6.0, 7.0, 8.0],
            [9.0, 10.0, 11.0, 12.0],
            [13.0, 14.0, 15.0, 16.0],
        ]

        mask = aa.Mask2D.manual(
            mask=np.array(
                [
                    [True, True, True, False],
                    [True, False, False, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scales=(1.0, 2.0),
            sub_size=1,
        )

        arr_masked = aa.Array2D.manual_mask(array=array_2d, mask=mask)

        extent = arr_masked.extent_of_zoomed_array(buffer=1)

        assert extent == pytest.approx(np.array([-4.0, 6.0, -2.0, 3.0]), 1.0e-4)

    def test__binned_up__compare_all_extract_methods_to_array_util(self):
        array_2d = np.array(
            [
                [1.0, 6.0, 3.0, 7.0, 3.0, 2.0],
                [2.0, 5.0, 3.0, 7.0, 7.0, 7.0],
                [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            ]
        )

        arr = aa.Array2D.manual_native(
            array=array_2d, sub_size=1, pixel_scales=(0.1, 0.1)
        )

        arr_binned_util = aa.util.binning.bin_array_2d_via_mean(
            array_2d=array_2d, bin_up_factor=4
        )
        arr_binned = arr.binned_up_from(bin_up_factor=4, method="mean")
        assert (arr_binned.native == arr_binned_util).all()
        assert arr_binned.pixel_scales == (0.4, 0.4)

        arr_binned_util = aa.util.binning.bin_array_2d_via_quadrature(
            array_2d=array_2d, bin_up_factor=4
        )
        arr_binned = arr.binned_up_from(bin_up_factor=4, method="quadrature")
        assert (arr_binned.native == arr_binned_util).all()
        assert arr_binned.pixel_scales == (0.4, 0.4)

        arr_binned_util = aa.util.binning.bin_array_2d_via_sum(
            array_2d=array_2d, bin_up_factor=4
        )
        arr_binned = arr.binned_up_from(bin_up_factor=4, method="sum")
        assert (arr_binned.native == arr_binned_util).all()
        assert arr_binned.pixel_scales == (0.4, 0.4)

    def test__binned_up__invalid_method__raises_exception(self):
        array_2d = [
            [1.0, 6.0, 3.0, 7.0, 3.0, 2.0],
            [2.0, 5.0, 3.0, 7.0, 7.0, 7.0],
            [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        ]

        array_2d = aa.Array2D.manual_native(
            array=array_2d, sub_size=1, pixel_scales=(0.1, 0.1)
        )
        with pytest.raises(exc.ArrayException):
            array_2d.binned_up_from(bin_up_factor=4, method="wrong")


class TestOutputToFits:
    def test__output_to_fits(self):

        arr = aa.Array2D.from_fits(
            file_path=path.join(test_data_dir, "3x3_ones.fits"), hdu=0, pixel_scales=1.0
        )

        output_data_dir = path.join(
            "{}".format(path.dirname(path.realpath(__file__))),
            "files",
            "array",
            "output_test",
        )
        if path.exists(output_data_dir):
            shutil.rmtree(output_data_dir)

        os.makedirs(output_data_dir)

        arr.output_to_fits(file_path=path.join(output_data_dir, "array.fits"))

        array_from_out = aa.Array2D.from_fits(
            file_path=path.join(output_data_dir, "array.fits"), hdu=0, pixel_scales=1.0
        )

        assert (array_from_out.native == np.ones((3, 3))).all()

    def test__output_to_fits__shapes_of_arrays_are_2d(self):

        arr = aa.Array2D.from_fits(
            file_path=path.join(test_data_dir, "3x3_ones.fits"), hdu=0, pixel_scales=1.0
        )

        output_data_dir = path.join(
            "{}".format(path.dirname(path.realpath(__file__))),
            "files",
            "array",
            "output_test",
        )

        if path.exists(output_data_dir):
            shutil.rmtree(output_data_dir)

        os.makedirs(output_data_dir)

        arr.output_to_fits(file_path=path.join(output_data_dir, "array.fits"))

        array_from_out = aa.util.array.numpy_array_2d_from_fits(
            file_path=path.join(output_data_dir, "array.fits"), hdu=0
        )

        assert (array_from_out == np.ones((3, 3))).all()

        mask = aa.Mask2D.unmasked(shape_native=(3, 3), pixel_scales=0.1)

        masked_array = aa.Array2D.manual_mask(array=arr, mask=mask)

        masked_array.output_to_fits(
            file_path=path.join(output_data_dir, "masked_array.fits")
        )

        masked_array_from_out = aa.util.array.numpy_array_2d_from_fits(
            file_path=path.join(output_data_dir, "masked_array.fits"), hdu=0
        )

        assert (masked_array_from_out == np.ones((3, 3))).all()


class TestExposureInfo:
    def test__exposure_info_has_date_and_time_of_observation__calcs_julian_date(self):

        exposure_info = aa.ExposureInfo(
            date_of_observation="2000-01-01", time_of_observation="00:00:00"
        )

        assert exposure_info.modified_julian_date == 51544.0



import numpy as np

from autoarray.structures import abstract_structure
from autoarray.util import fit_util


class FitData:
    def __init__(
        self,
        data,
        noise_map,
        model_data,
        mask=None,
        inversion=None,
        use_mask_in_fit=False,
    ):

        self.data = data
        self.noise_map = noise_map
        self.model_data = model_data
        self._mask = mask
        self.inversion = inversion
        self.use_mask_in_fit = use_mask_in_fit

    @property
    def mask(self):
        return self._mask

    @property
    def residual_map(self) -> abstract_structure.AbstractStructure:
        """
        Returns the residual-map between the masked dataset and model data, where:

        Residuals = (Data - Model_Data).
        """
        if self.use_mask_in_fit:
            return fit_util.residual_map_with_mask_from(
                data=self.data, model_data=self.model_data, mask=self.mask
            )
        return fit_util.residual_map_from(data=self.data, model_data=self.model_data)

    @property
    def normalized_residual_map(self) -> abstract_structure.AbstractStructure:
        """
        Returns the normalized residual-map between the masked dataset and model data, where:

        Normalized_Residual = (Data - Model_Data) / Noise
        """
        if self.use_mask_in_fit:
            return fit_util.normalized_residual_map_with_mask_from(
                residual_map=self.residual_map, noise_map=self.noise_map, mask=self.mask
            )
        return fit_util.normalized_residual_map_from(
            residual_map=self.residual_map, noise_map=self.noise_map
        )

    @property
    def chi_squared_map(self) -> abstract_structure.AbstractStructure:
        """
        Returns the chi-squared-map between the residual-map and noise-map, where:

        Chi_Squared = ((Residuals) / (Noise)) ** 2.0 = ((Data - Model)**2.0)/(Variances)
        """
        if self.use_mask_in_fit:
            return fit_util.chi_squared_map_with_mask_from(
                residual_map=self.residual_map, noise_map=self.noise_map, mask=self.mask
            )
        return fit_util.chi_squared_map_from(
            residual_map=self.residual_map, noise_map=self.noise_map
        )

    @property
    def signal_to_noise_map(self) -> abstract_structure.AbstractStructure:
        """The signal-to-noise_map of the dataset and noise-map which are fitted."""
        signal_to_noise_map = np.divide(self.data, self.noise_map)
        signal_to_noise_map[signal_to_noise_map < 0] = 0
        return signal_to_noise_map

    @property
    def chi_squared(self) -> float:
        """
        Returns the chi-squared terms of the model data's fit to an dataset, by summing the chi-squared-map.
        """
        if self.use_mask_in_fit:
            return fit_util.chi_squared_with_mask_from(
                chi_squared_map=self.chi_squared_map, mask=self.mask
            )
        return fit_util.chi_squared_from(chi_squared_map=self.chi_squared_map)

    @property
    def reduced_chi_squared(self) -> float:
        return self.chi_squared / int(np.size(self.mask) - np.sum(self.mask))

    @property
    def noise_normalization(self) -> float:
        """
        Returns the noise-map normalization term of the noise-map, summing the noise_map value in every pixel as:

        [Noise_Term] = sum(log(2*pi*[Noise]**2.0))
        """
        if self.use_mask_in_fit:
            return fit_util.noise_normalization_with_mask_from(
                noise_map=self.noise_map, mask=self.mask
            )
        return fit_util.noise_normalization_from(noise_map=self.noise_map)

    @property
    def log_likelihood(self) -> float:
        """
        Returns the log likelihood of each model data point's fit to the dataset, where:

        Log Likelihood = -0.5*[Chi_Squared_Term + Noise_Term] (see functions above for these definitions)
        """
        return fit_util.log_likelihood_from(
            chi_squared=self.chi_squared, noise_normalization=self.noise_normalization
        )

    @property
    def log_likelihood_with_regularization(self) -> float:
        """
        Returns the log likelihood of an inversion's fit to the dataset, including a regularization term which \
        comes from an inversion:

        Log Likelihood = -0.5*[Chi_Squared_Term + Regularization_Term + Noise_Term] (see functions above for these definitions)
        """
        if self.inversion is not None:
            return fit_util.log_likelihood_with_regularization_from(
                chi_squared=self.chi_squared,
                regularization_term=self.inversion.regularization_term,
                noise_normalization=self.noise_normalization,
            )

    @property
    def log_evidence(self) -> float:
        """
        Returns the log evidence of the inversion's fit to a dataset, where the log evidence includes a number of terms
        which quantify the complexity of an inversion's reconstruction (see the `Inversion` module):

        Log Evidence = -0.5*[Chi_Squared_Term + Regularization_Term + Log(Covariance_Regularization_Term) -
                           Log(Regularization_Matrix_Term) + Noise_Term]

        Parameters
        ----------
        chi_squared : float
            The chi-squared term of the inversion's fit to the data.
        regularization_term : float
            The regularization term of the inversion, which is the sum of the difference between reconstructed \
            flux of every pixel multiplied by the regularization coefficient.
        log_curvature_regularization_term : float
            The log of the determinant of the sum of the curvature and regularization matrices.
        log_regularization_term : float
            The log of the determinant o the regularization matrix.
        noise_normalization : float
            The normalization noise_map-term for the data's noise-map.
        """
        if self.inversion is not None:
            return fit_util.log_evidence_from(
                chi_squared=self.chi_squared,
                regularization_term=self.inversion.regularization_term,
                log_curvature_regularization_term=self.inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=self.inversion.log_det_regularization_matrix_term,
                noise_normalization=self.noise_normalization,
            )

    @property
    def figure_of_merit(self) -> float:
        if self.inversion is None:
            return self.log_likelihood
        else:
            return self.log_evidence

    @property
    def total_inversions(self) -> int:
        if self.inversion is None:
            return 0
        else:
            return 1


class FitDataset(FitData):

    # noinspection PyUnresolvedReferences
    def __init__(
        self, masked_dataset, model_data, inversion=None, use_mask_in_fit=True
    ):
        """Class to fit a masked dataset where the dataset's data structures are any dimension.

        Parameters
        -----------
        masked_dataset : MaskedDataset
            The masked dataset (data, mask, noise-map, etc.) that is fitted.
        model_data : np.ndarray
            The model data the masked dataset is fitted with.
        inversion : Inversion
            If the fit uses an `Inversion` this is the instance of the object used to perform the fit. This determines
            if the `log_likelihood` or `log_evidence` is used as the `figure_of_merit`.
        use_mask_in_fit : bool
            If `True`, masked data points are omitted from the fit. If `False` they are not (in most use cases the
            `masked_dataset` will have been processed to remove masked points, for example the `slim` representation).

        Attributes
        -----------
        residual_map : np.ndarray
            The residual-map of the fit (data - model_data).
        chi_squared_map : np.ndarray
            The chi-squared-map of the fit ((data - model_data) / noise_maps ) **2.0
        chi_squared : float
            The overall chi-squared of the model's fit to the dataset, summed over every data point.
        reduced_chi_squared : float
            The reduced chi-squared of the model's fit to simulate (chi_squared / number of data points), summed over \
            every data point.
        noise_normalization : float
            The overall normalization term of the noise_map, summed over every data point.
        log_likelihood : float
            The overall log likelihood of the model's fit to the dataset, summed over evey data point.
        """

        self.masked_dataset = masked_dataset

        super().__init__(
            data=masked_dataset.data,
            noise_map=masked_dataset.noise_map,
            model_data=model_data,
            mask=masked_dataset.mask,
            inversion=inversion,
        )

        self.use_mask_in_fit = use_mask_in_fit

    @property
    def name(self) -> str:
        return self.masked_dataset.dataset.name


class FitImaging(FitDataset):
    def __init__(
        self, masked_imaging, model_image, inversion=None, use_mask_in_fit=True
    ):
        """Class to fit a masked imaging dataset.

        Parameters
        -----------
        masked_imaging : MaskedImaging
            The masked imaging dataset that is fitted.
        model_image : Array2D
            The model image the masked imaging is fitted with.
        inversion : Inversion
            If the fit uses an `Inversion` this is the instance of the object used to perform the fit. This determines
            if the `log_likelihood` or `log_evidence` is used as the `figure_of_merit`.
        use_mask_in_fit : bool
            If `True`, masked data points are omitted from the fit. If `False` they are not (in most use cases the
            `masked_dataset` will have been processed to remove masked points, for example the `slim` representation).

        Attributes
        -----------
        residual_map : np.ndarray
            The residual-map of the fit (data - model_data).
        chi_squared_map : np.ndarray
            The chi-squared-map of the fit ((data - model_data) / noise_maps ) **2.0
        chi_squared : float
            The overall chi-squared of the model's fit to the dataset, summed over every data point.
        reduced_chi_squared : float
            The reduced chi-squared of the model's fit to simulate (chi_squared / number of data points), summed over \
            every data point.
        noise_normalization : float
            The overall normalization term of the noise_map, summed over every data point.
        log_likelihood : float
            The overall log likelihood of the model's fit to the dataset, summed over evey data point.
        """

        super().__init__(
            masked_dataset=masked_imaging,
            model_data=model_image,
            inversion=inversion,
            use_mask_in_fit=use_mask_in_fit,
        )

    @property
    def masked_imaging(self):
        return self.masked_dataset

    @property
    def image(self) -> abstract_structure.AbstractStructure:
        return self.data

    @property
    def model_image(self) -> abstract_structure.AbstractStructure:
        return self.model_data


class FitInterferometer(FitDataset):
    def __init__(
        self,
        masked_interferometer,
        model_visibilities,
        inversion=None,
        use_mask_in_fit=True,
    ):
        """Class to fit a masked interferometer dataset.

        Parameters
        -----------
        masked_interferometer : MaskedInterferometer
            The masked interferometer dataset that is fitted.
        model_visibilities : Visibilities
            The model visibilities the masked imaging is fitted with.
        inversion : Inversion
            If the fit uses an `Inversion` this is the instance of the object used to perform the fit. This determines
            if the `log_likelihood` or `log_evidence` is used as the `figure_of_merit`.
        use_mask_in_fit : bool
            If `True`, masked data points are omitted from the fit. If `False` they are not (in most use cases the
            `masked_dataset` will have been processed to remove masked points, for example the `slim` representation).

        Attributes
        -----------
        residual_map : np.ndarray
            The residual-map of the fit (data - model_data).
        chi_squared_map : np.ndarray
            The chi-squared-map of the fit ((data - model_data) / noise_maps ) **2.0
        chi_squared : float
            The overall chi-squared of the model's fit to the dataset, summed over every data point.
        reduced_chi_squared : float
            The reduced chi-squared of the model's fit to simulate (chi_squared / number of data points), summed over \
            every data point.
        noise_normalization : float
            The overall normalization term of the noise_map, summed over every data point.
        log_likelihood : float
            The overall log likelihood of the model's fit to the dataset, summed over evey data point.
        """

        super(FitInterferometer, self).__init__(
            masked_dataset=masked_interferometer,
            model_data=model_visibilities,
            inversion=inversion,
            use_mask_in_fit=use_mask_in_fit,
        )

    @property
    def masked_interferometer(self):
        return self.masked_dataset

    @property
    def mask(self) -> abstract_structure.AbstractStructure:
        return self.masked_interferometer.visibilities_mask

    @property
    def visibilities_mask(self) -> abstract_structure.AbstractStructure:
        return self.masked_interferometer.visibilities_mask

    @property
    def visibilities(self) -> abstract_structure.AbstractStructure:
        return self.data

    @property
    def signal_to_noise_map(self) -> abstract_structure.AbstractStructure:
        """The signal-to-noise_map of the dataset and noise-map which are fitted."""
        signal_to_noise_map_real = np.divide(
            np.real(self.data), np.real(self.noise_map)
        )
        signal_to_noise_map_real[signal_to_noise_map_real < 0] = 0.0
        signal_to_noise_map_imag = np.divide(
            np.imag(self.data), np.imag(self.noise_map)
        )
        signal_to_noise_map_imag[signal_to_noise_map_imag < 0] = 0.0

        return signal_to_noise_map_real + 1.0j * signal_to_noise_map_imag

    @property
    def model_visibilities(self) -> abstract_structure.AbstractStructure:
        return self.model_data

    @property
    def normalized_residual_map(self) -> abstract_structure.AbstractStructure:
        """
        Returns the normalized residual-map between the masked dataset and model data, where:

        Normalized_Residual = (Data - Model_Data) / Noise
        """
        return fit_util.normalized_residual_map_complex_with_mask_from(
            residual_map=self.residual_map, noise_map=self.noise_map, mask=self.mask
        )

    @property
    def chi_squared_map(self) -> abstract_structure.AbstractStructure:
        """
        Returns the chi-squared-map between the residual-map and noise-map, where:

        Chi_Squared = ((Residuals) / (Noise)) ** 2.0 = ((Data - Model)**2.0)/(Variances)
        """
        return fit_util.chi_squared_map_complex_with_mask_from(
            residual_map=self.residual_map, noise_map=self.noise_map, mask=self.mask
        )

    @property
    def chi_squared(self) -> float:
        """
        Returns the chi-squared terms of the model data's fit to an dataset, by summing the chi-squared-map.
        """
        return fit_util.chi_squared_complex_with_mask_from(
            chi_squared_map=self.chi_squared_map, mask=self.mask
        )

    @property
    def noise_normalization(self) -> float:
        """
        Returns the noise-map normalization term of the noise-map, summing the noise_map value in every pixel as:

        [Noise_Term] = sum(log(2*pi*[Noise]**2.0))
        """
        return fit_util.noise_normalization_complex_with_mask_from(
            noise_map=self.noise_map, mask=self.mask
        )



import logging

import numpy as np
import pylops

from autoarray.mask import mask_2d
from autoarray.structures import grids
from autoarray.util import array_util

logging.basicConfig()
logger = logging.getLogger(__name__)


class AbstractVisibilities(np.ndarray):

    # noinspection PyUnusedLocal
    def __new__(cls, visibilities, *args, **kwargs):
        """
        A collection of (real, imag) visibilities which are used to represent the data in an `Interferometer` dataset.

        The (real, imag) visibilities are stored as a 1D complex NumPy array of shape [total_visibilities]. These can
        be mapped to a 2D real NumPy array of shape [total_visibilities, 2] and a `Grid2DIrregular` data structure
        which is used for plotting the visibilities in 2D in the complex plane.

        Calculations should use the NumPy array structure wherever possible for efficient calculations.

        The vectors input to this function can have any of the following forms (they will be converted to the 1D
        complex NumPy array structure and can be converted back using the object's properties):

        [1.0+1.0j, 2.0+2.0j]
        [[1.0, 1.0], [2.0, 2.0]]

        Parameters
        ----------
        visibilities : np.ndarray
            The (real, imag) visibilities values.
        """

        if type(visibilities) is list:
            visibilities = np.asarray(visibilities)

        if "float" in str(visibilities.dtype):
            if visibilities.shape[1] == 2:
                visibilities = (
                    np.apply_along_axis(lambda args: [complex(*args)], 1, visibilities)
                    .astype("complex128")
                    .ravel()
                )

        return visibilities.view(cls)

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(AbstractVisibilities, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return pickled_state[0], pickled_state[1], new_state

    # noinspection PyMethodOverriding
    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super(AbstractVisibilities, self).__setstate__(state[0:-1])

    @property
    def slim(self):
        return self

    @property
    def in_array(self):
        """
        Returns the 1D complex NumPy array of values with shape [total_visibilities] as a NumPy float array of
        shape [total_visibilities, 2].
        """
        return np.stack((np.real(self), np.imag(self)), axis=-1)

    @property
    def in_grid(self):
        """Returns the 1D complex NumPy array of values as an irregular grid."""
        return grids.Grid2DIrregular(grid=self.in_array)

    @property
    def shape_slim(self):
        return self.shape[0]

    @property
    @array_util.Memoizer()
    def amplitudes(self):
        return np.sqrt(np.square(self.real) + np.square(self.imag))

    @property
    @array_util.Memoizer()
    def phases(self):
        return np.arctan2(self.imag, self.real)

    def output_to_fits(self, file_path, overwrite=False):
        """
        Output the visibilities to a .fits file.

        The complex values are converted to a 2D NumPy float array of shape [total_visiblities, 2] before being
        written to `.fits` format via the `in_array` property.

        Parameters
        ----------
        file_path : str
            The path the file is output to, including the filename and the ``.fits`` extension,
            e.g. '/path/to/filename.fits'
        overwrite : bool
            If a file already exists at the path, if overwrite=True it is overwritten else an error is raised.
        """
        array_util.numpy_array_2d_to_fits(
            array_2d=self.in_array, file_path=file_path, overwrite=overwrite
        )

    @property
    def scaled_maxima(self):
        """The maximum values of the visibilities if they are treated as a 2D grid in the complex plane."""
        return (np.max(self.real), np.max(self.imag))

    @property
    def scaled_minima(self):
        """The minimum values of the visibilities if they are treated as a 2D grid in the complex plane."""
        return (np.min(self.real), np.min(self.imag))

    @property
    def extent(self):
        """The extent of the visibilities if they are treated as a 2D grid in the complex plane."""
        return np.asarray(
            [
                self.scaled_minima[1],
                self.scaled_maxima[1],
                self.scaled_minima[0],
                self.scaled_maxima[0],
            ]
        )


class Visibilities(AbstractVisibilities):
    @classmethod
    def manual_slim(cls, visibilities):
        """
        Create `Visibilities` (see `AbstractVisibilities.__new__`) by inputting (real, imag) values as a 1D complex
        NumPy array or 2D NumPy float array or list, for example:

        visibilities=np.array([1.0+1.0j, 2.0+2.0j, 3.0+3.0j, 4.0+4.0j])
        visibilities=np.array([[1.0+1.0], [2.0+2.0], [3.0+3.0], [4.0+4.0]])
        visibilities=[[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [4.0, 4.0]]

        Parameters
        ----------
        visibilities : np.ndarray
            The (real, imag) visibilities values.
        """
        return cls(visibilities=visibilities)

    @classmethod
    def full(cls, fill_value, shape_slim):
        """
        Create `Visibilities` (see `AbstractVisibilities.__new__`) where all (real, imag) values are filled with an
        input fill value, analogous to the method numpy ndarray.full.

        From 1D input the method cannot determine the 2D shape of the array and its mask, thus the shape_native must be
        input into this method. The mask is setup as a unmasked `Mask2D` of shape_native.

        Parameters
        ----------
        fill_value : float
            The value all real and imaginary array elements are filled with.
        shape_slim : int
            The 1D shape of output visibilities.
        """
        return cls.manual_slim(
            visibilities=np.full(
                fill_value=fill_value + fill_value * 1j, shape=(shape_slim[0],)
            )
        )

    @classmethod
    def ones(cls, shape_slim):
        """
        Create `Visibilities` (see `AbstractVisibilities.__new__`) where all (real, imag) values are filled with ones,
        analogous to the method np.ones().

        From 1D input the method cannot determine the 2D shape of the array and its mask, thus the shape_native must be
        input into this method. The mask is setup as a unmasked `Mask2D` of shape_native.

        Parameters
        ----------
        shape_slim : int
            The 1D shape of output visibilities.
        """
        return cls.full(fill_value=1.0, shape_slim=shape_slim)

    @classmethod
    def zeros(cls, shape_slim):
        """
        Create `Visibilities` (see `AbstractVisibilities.__new__`) where all (real, imag) values are filled with zeros,
        analogous to the method np.zeros().

        From 1D input the method cannot determine the 2D shape of the array and its mask, thus the shape_native must be
        input into this method. The mask is setup as a unmasked `Mask2D` of shape_native.

        Parameters
        ----------
        shape_slim : int
            The 1D shape of output visibilities.
        """
        return cls.full(fill_value=0.0, shape_slim=shape_slim)

    @classmethod
    def from_fits(cls, file_path, hdu):
        """
        Create `Visibilities` (see `AbstractVisibilities.__new__`) by loading the(real, imag) values from a .fits file.

        The `.fits` file stores these values as a real set of values of shape [total_visibilities, 2] which are
        converted to a 1d complex NumPy array.

        Parameters
        ----------
        file_path : str
            The path the file is loaded from, including the filename and the ``.fits`` extension,
            e.g. '/path/to/filename.fits'
        hdu : int
            The Header-Data Unit of the .fits file the visibilitiy data is loaded from.
        """
        visibilities_1d = array_util.numpy_array_2d_from_fits(
            file_path=file_path, hdu=hdu
        )
        return cls.manual_slim(visibilities=visibilities_1d)


class VisibilitiesNoiseMap(Visibilities):

    # noinspection PyUnusedLocal
    def __new__(cls, visibilities, *args, **kwargs):
        """
        A collection of (real, imag) visibilities noise-map values which are used to represent the noise-map in
        an `Interferometer` dataset.

        This data structure behaves the same as the `Visibilities` structure (see `AbstractVisibilities.__new__`). The
        only difference is that it includes a `WeightOperator` used by `Inversion`'s which use `LinearOperators` and
        the library `PyLops` to fit `Interferometer` data.

        Parameters
        ----------
        visibilities : np.ndarray
            The (real, imag) visibilities values.
        """

        if type(visibilities) is list:
            visibilities = np.asarray(visibilities)

        if "float" in str(visibilities.dtype):
            if visibilities.shape[1] == 2:
                visibilities = (
                    np.apply_along_axis(lambda args: [complex(*args)], 1, visibilities)
                    .astype("complex128")
                    .ravel()
                )

        obj = super(VisibilitiesNoiseMap, cls).__new__(
            cls=cls, visibilities=visibilities
        )

        weights = np.reciprocal(np.real(visibilities)) + 1j * np.reciprocal(
            np.imag(visibilities)
        )

        obj.Wop = pylops.Diagonal(np.real(weights.ravel()), dtype="complex128")

        return obj

    def __array_finalize__(self, obj):

        if hasattr(obj, "Wop"):
            self.Wop = obj.Wop



import matplotlib

from autoconf import conf

backend = conf.get_matplotlib_backend()

if backend not in "default":
    matplotlib.use(backend)

import numpy as np

from autoarray.fit.fit import FitDataset


def get_real_space_mask_from(mask: np.ndarray, fit: FitDataset) -> np.ndarray:
    """Get the masks of the fit if the masks should be plotted on the fit.

    Parameters
    -----------
    fit : FitDataset
        The fit to the datas, which includes a lisrt of every model image, residual_map, chi-squareds, etc.
    mask : bool
        If `True`, the masks is plotted on the fit's datas.
    """
    if mask:
        return fit.masked_dataset.settings.mask


def radii_bin_size_from_minimum_and_maximum_radii_and_radii_points(
    minimum_radius, maximum_radius, radii_points
):
    return (maximum_radius - minimum_radius) / radii_points


def quantity_radii_from_minimum_and_maximum_radii_and_radii_points(
    minimum_radius, maximum_radius, radii_points
):
    return list(
        np.linspace(start=minimum_radius, stop=maximum_radius, num=radii_points + 1)
    )


def quantity_and_annuli_radii_from_minimum_and_maximum_radii_and_radii_points(
    minimum_radius, maximum_radius, radii_points
):
    radii_bin_size = radii_bin_size_from_minimum_and_maximum_radii_and_radii_points(
        minimum_radius=minimum_radius,
        maximum_radius=maximum_radius,
        radii_points=radii_points,
    )

    quantity_radii = list(
        np.linspace(
            start=minimum_radius + radii_bin_size / 2.0,
            stop=maximum_radius - radii_bin_size / 2.0,
            num=radii_points,
        )
    )
    annuli_radii = list(
        np.linspace(start=minimum_radius, stop=maximum_radius, num=radii_points + 1)
    )

    return quantity_radii, annuli_radii



from .wrap_base import (
    Units,
    Figure,
    Axis,
    Cmap,
    Colorbar,
    ColorbarTickParams,
    TickParams,
    YTicks,
    XTicks,
    Title,
    YLabel,
    XLabel,
    Legend,
    Output,
)
from .wrap_1d import LinePlot, AXVLine
from .wrap_2d import (
    ArrayOverlay,
    GridScatter,
    GridPlot,
    VectorFieldQuiver,
    PatchOverlay,
    VoronoiDrawer,
    OriginScatter,
    MaskScatter,
    BorderScatter,
    PositionsScatter,
    IndexScatter,
    PixelizationGridScatter,
    ParallelOverscanPlot,
    SerialPrescanPlot,
    SerialOverscanPlot,
)



from autoconf import conf
import typing


class AbstractInclude:
    def __init__(
        self, origin: typing.Optional[bool] = None, mask: typing.Optional[bool] = None
    ):
        """
        Sets which `Visuals` are included on a figure that is plotted using a `Plotter`.

        The `Include` object is used to extract the visuals of the plotted data structure (e.g. `Array2D`, `Grid2D`) so
        they can be used in plot functions. Only visuals with a `True` entry in the `Include` object are extracted and t
        plotted.

        If an entry is not input into the class (e.g. it retains its default entry of `None`) then the bool is
        loaded from the `config/visualize/include.ini` config file. This means the default visuals of a project
        can be specified in a config file.

        Parameters
        ----------
        origin : bool
            If `True`, the `origin` of the plotted data structure (e.g. `Array2D`, `Grid2D`)  is included on the figure.
        mask : bool
            if `True`, the `mask` of the plotted data structure (e.g. `Array2D`, `Grid2D`)  is included on the figure.
        """

        self._origin = origin
        self._mask = mask

    def load(self, value, name):
        if value is True:
            return True
        elif value is False:
            return False
        elif value is None:
            return conf.instance["visualize"]["include"]["include"][name]

    @property
    def origin(self):
        return self.load(value=self._origin, name="origin")

    @property
    def mask(self):
        return self.load(value=self._mask, name="mask")


class Include1D(AbstractInclude):
    def __init__(
        self, origin: typing.Optional[bool] = None, mask: typing.Optional[bool] = None
    ):
        """
        Sets which `Visuals1D` are included on a figure plotting 1D data that is plotted using a `Plotter1D`.

        The `Include` object is used to extract the visuals of the plotted 1D data structures so they can be used in 
        plot functions. Only visuals with a `True` entry in the `Include` object are extracted and plotted.

        If an entry is not input into the class (e.g. it retains its default entry of `None`) then the bool is
        loaded from the `config/visualize/include.ini` config file. This means the default visuals of a project
        can be specified in a config file.

        Parameters
        ----------
        origin : bool
            If `True`, the `origin` of the plotted data structure (e.g. `Line`)  is included on the figure.
        mask : bool
            if `True`, the `mask` of the plotted data structure (e.g. `Line`)  is included on the figure.
        """
        super().__init__(origin=origin, mask=mask)


class Include2D(AbstractInclude):
    def __init__(
        self,
        origin: typing.Optional[bool] = None,
        mask: typing.Optional[bool] = None,
        border: typing.Optional[bool] = None,
        grid: typing.Optional[bool] = None,
        mapper_data_pixelization_grid: typing.Optional[bool] = None,
        mapper_source_pixelization_grid: typing.Optional[bool] = None,
        mapper_source_grid_slim: typing.Optional[bool] = None,
        parallel_overscan: typing.Optional[bool] = None,
        serial_prescan: typing.Optional[bool] = None,
        serial_overscan: typing.Optional[bool] = None,
    ):
        """
        Sets which `Visuals2D` are included on a figure plotting 2D data that is plotted using a `Plotter2D`.

        The `Include` object is used to extract the visuals of the plotted 2D data structures so they can be used in 
        plot functions. Only visuals with a `True` entry in the `Include` object are extracted and plotted.

        If an entry is not input into the class (e.g. it retains its default entry of `None`) then the bool is
        loaded from the `config/visualize/include.ini` config file. This means the default visuals of a project
        can be specified in a config file.

        Parameters
        ----------
        origin : bool
            If `True`, the `origin` of the plotted data structure (e.g. `Array2D`, `Grid2D`)  is included on the figure.
        mask : bool
            if `True`, the `mask` of the plotted data structure (e.g. `Array2D`, `Grid2D`)  is included on the figure.
        border : bool
            If `True`, the `border` of the plotted data structure (e.g. `Array2D`, `Grid2D`)  is included on the figure.
        mapper_data_pixelization_grid : bool
            If `True`, the pixelization grid in the data plane of a plotted `Mapper` is included on the figure.
        mapper_source_pixelization_grid : bool
            If `True`, the pixelization grid in the source plane of a plotted `Mapper` is included on the figure.
        parallel_overscan : bool
            If `True`, the parallel overscan of a plotted `Frame2D` is included on the figure.
        serial_prescan : bool
            If `True`, the serial prescan of a plotted `Frame2D` is included on the figure.
        serial_overscan : bool
            If `True`, the serial overscan of a plotted `Frame2D` is included on the figure.
        """

        super().__init__(origin=origin, mask=mask)

        self._border = border
        self._grid = grid
        self._mapper_data_pixelization_grid = mapper_data_pixelization_grid
        self._mapper_source_pixelization_grid = mapper_source_pixelization_grid
        self._mapper_source_grid_slim = mapper_source_grid_slim
        self._parallel_overscan = parallel_overscan
        self._serial_prescan = serial_prescan
        self._serial_overscan = serial_overscan

    @property
    def border(self):
        return self.load(value=self._border, name="border")

    @property
    def grid(self):
        return self.load(value=self._grid, name="grid")

    @property
    def mapper_data_pixelization_grid(self):
        return self.load(
            value=self._mapper_data_pixelization_grid,
            name="mapper_data_pixelization_grid",
        )

    @property
    def mapper_source_pixelization_grid(self):
        return self.load(
            value=self._mapper_source_pixelization_grid,
            name="mapper_source_pixelization_grid",
        )

    @property
    def mapper_source_grid_slim(self):
        return self.load(
            value=self._mapper_source_grid_slim, name="mapper_source_grid_slim"
        )

    @property
    def parallel_overscan(self):
        return self.load(value=self._parallel_overscan, name="parallel_overscan")

    @property
    def serial_prescan(self):
        return self.load(value=self._serial_prescan, name="serial_prescan")

    @property
    def serial_overscan(self):
        return self.load(value=self._serial_overscan, name="serial_overscan")



import os
from os import path

import numpy as np
import pytest

import autoarray as aa
from autoarray import exc

test_data_dir = path.join(
    "{}".format(os.path.dirname(os.path.realpath(__file__))), "files", "array"
)


class TestAPI:
    def test__manual__makes_array_without_other_inputs(self):

        arr = aa.Array2D.manual_native(array=[[1.0, 2.0], [3.0, 4.0]], pixel_scales=1.0)

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 0.0)

        arr = aa.Array2D.manual_slim(
            array=[1.0, 2.0, 3.0, 4.0],
            shape_native=(2, 2),
            pixel_scales=1.0,
            origin=(0.0, 1.0),
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

        arr = aa.Array2D.manual_slim(
            array=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            shape_native=(2, 3),
            pixel_scales=(2.0, 3.0),
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])).all()
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])).all()
        assert arr.pixel_scales == (2.0, 3.0)
        assert arr.origin == (0.0, 0.0)

    def test__manual__makes_with_sub_size(self):

        arr = aa.Array2D.manual_native(
            array=[[1.0, 2.0], [3.0, 4.0]], pixel_scales=1.0, sub_size=1
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 0.0)
        assert arr.mask.sub_size == 1

        arr = aa.Array2D.manual_slim(
            array=[1.0, 2.0, 3.0, 4.0],
            shape_native=(1, 1),
            pixel_scales=1.0,
            sub_size=2,
            origin=(0.0, 1.0),
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)
        assert arr.mask.sub_size == 2

        arr = aa.Array2D.manual_slim(
            array=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0],
            shape_native=(2, 1),
            pixel_scales=2.0,
            sub_size=2,
        )

        assert type(arr) == aa.Array2D
        assert (
            arr.native == np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])
        ).all()
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])).all()
        assert (arr.native_binned == np.array([[2.5], [6.5]])).all()
        assert (arr.slim_binned == np.array([2.5, 6.5])).all()
        assert arr.pixel_scales == (2.0, 2.0)
        assert arr.origin == (0.0, 0.0)
        assert arr.mask.sub_size == 2

    def test__manual_mask__makes_array_without_other_inputs(self):
        mask = aa.Mask2D.unmasked(shape_native=(2, 2), pixel_scales=1.0)
        arr = aa.Array2D.manual_mask(array=[[1.0, 2.0], [3.0, 4.0]], mask=mask)

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 0.0)

        mask = aa.Mask2D.manual(
            mask=[[False, False], [True, False]], pixel_scales=1.0, origin=(0.0, 1.0)
        )
        arr = aa.Array2D.manual_mask(array=[1.0, 2.0, 4.0], mask=mask)

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 2.0], [0.0, 4.0]])).all()
        assert (arr.slim == np.array([1.0, 2.0, 4.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

        mask = aa.Mask2D.manual(
            mask=[[False, False], [True, False]], pixel_scales=1.0, origin=(0.0, 1.0)
        )
        arr = aa.Array2D.manual_mask(array=[[1.0, 2.0], [3.0, 4.0]], mask=mask)

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 2.0], [0.0, 4.0]])).all()
        assert (arr.slim == np.array([1.0, 2.0, 4.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

        mask = aa.Mask2D.manual(mask=[[False], [True]], pixel_scales=2.0, sub_size=2)
        arr = aa.Array2D.manual_mask(
            array=[1.0, 2.0, 3.0, 4.0], mask=mask, store_slim=True
        )

        assert type(arr) == aa.Array2D
        assert (arr == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (
            arr.native == np.array([[1.0, 2.0], [3.0, 4.0], [0.0, 0.0], [0.0, 0.0]])
        ).all()
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (arr.native_binned == np.array([[2.5], [0.0]])).all()
        assert (arr.slim_binned == np.array([2.5])).all()
        assert arr.pixel_scales == (2.0, 2.0)
        assert arr.origin == (0.0, 0.0)
        assert arr.mask.sub_size == 2

        arr = aa.Array2D.manual_mask(
            array=[1.0, 2.0, 3.0, 4.0], mask=mask, store_slim=False
        )

        assert type(arr) == aa.Array2D
        assert (arr == np.array([[1.0, 2.0], [3.0, 4.0], [0.0, 0.0], [0.0, 0.0]])).all()
        assert (
            arr.native == np.array([[1.0, 2.0], [3.0, 4.0], [0.0, 0.0], [0.0, 0.0]])
        ).all()
        assert (arr.slim == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (arr.native_binned == np.array([[2.5], [0.0]])).all()
        assert (arr.slim_binned == np.array([2.5])).all()
        assert arr.pixel_scales == (2.0, 2.0)
        assert arr.origin == (0.0, 0.0)
        assert arr.mask.sub_size == 2

    def test__manual_native__exception_raised_if_input_array_is_2d_and_not_sub_shape_of_mask(
        self,
    ):
        with pytest.raises(exc.ArrayException):
            mask = aa.Mask2D.unmasked(shape_native=(2, 2), pixel_scales=1.0, sub_size=1)
            aa.Array2D.manual_mask(array=[[1.0], [3.0]], mask=mask)

        with pytest.raises(exc.ArrayException):
            mask = aa.Mask2D.unmasked(shape_native=(2, 2), pixel_scales=1.0, sub_size=2)
            aa.Array2D.manual_mask(array=[[1.0, 2.0], [3.0, 4.0]], mask=mask)

        with pytest.raises(exc.ArrayException):
            mask = aa.Mask2D.unmasked(shape_native=(2, 2), pixel_scales=1.0, sub_size=2)
            aa.Array2D.manual_mask(
                array=[[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], mask=mask
            )

    def test__exception_raised_if_input_array_is_1d_and_not_number_of_masked_sub_pixels(
        self,
    ):
        with pytest.raises(exc.ArrayException):
            mask = aa.Mask2D.manual(
                mask=[[False, False], [True, False]], pixel_scales=1.0, sub_size=1
            )
            aa.Array2D.manual_mask(array=[1.0, 2.0, 3.0, 4.0], mask=mask)

        with pytest.raises(exc.ArrayException):
            mask = aa.Mask2D.manual(
                mask=[[False, False], [True, False]], pixel_scales=1.0, sub_size=1
            )
            aa.Array2D.manual_mask(array=[1.0, 2.0], mask=mask)

        with pytest.raises(exc.ArrayException):
            mask = aa.Mask2D.manual(
                mask=[[False, True], [True, True]], pixel_scales=1.0, sub_size=2
            )
            aa.Array2D.manual_mask(array=[1.0, 2.0, 4.0], mask=mask)

        with pytest.raises(exc.ArrayException):
            mask = aa.Mask2D.manual(
                mask=[[False, True], [True, True]], pixel_scales=1.0, sub_size=2
            )
            aa.Array2D.manual_mask(array=[1.0, 2.0, 3.0, 4.0, 5.0], mask=mask)

    def test__full__makes_scaled_array_with_pixel_scale(self):

        arr = aa.Array2D.ones(shape_native=(2, 2), pixel_scales=1.0)

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 1.0], [1.0, 1.0]])).all()
        assert (arr.slim == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 0.0)

        arr = aa.Array2D.full(
            fill_value=2.0, shape_native=(2, 2), pixel_scales=1.0, origin=(0.0, 1.0)
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[2.0, 2.0], [2.0, 2.0]])).all()
        assert (arr.slim == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

    def test__full__makes_scaled_sub_array_with_pixel_scale_and_sub_size(self):

        arr = aa.Array2D.full(
            fill_value=1.0, shape_native=(1, 4), pixel_scales=1.0, sub_size=1
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 1.0, 1.0, 1.0]])).all()
        assert (arr.slim == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 0.0)
        assert arr.mask.sub_size == 1

        arr = aa.Array2D.full(
            fill_value=2.0,
            shape_native=(1, 1),
            pixel_scales=1.0,
            sub_size=2,
            origin=(0.0, 1.0),
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[2.0, 2.0], [2.0, 2.0]])).all()
        assert (arr.slim == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)
        assert arr.mask.sub_size == 2

    def test__ones_zeros__makes_array_without_other_inputs(self):

        arr = aa.Array2D.ones(shape_native=(2, 2), pixel_scales=1.0)

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 1.0], [1.0, 1.0]])).all()
        assert (arr.slim == np.array([1.0, 1.0, 1.0, 1.0])).all()

        arr = aa.Array2D.zeros(shape_native=(2, 2), pixel_scales=1.0, store_slim=True)

        assert type(arr) == aa.Array2D
        assert (arr == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert (arr.native == np.array([[0.0, 0.0], [0.0, 0.0]])).all()
        assert (arr.slim == np.array([0.0, 0.0, 0.0, 0.0])).all()

        arr = aa.Array2D.zeros(shape_native=(2, 2), pixel_scales=1.0, store_slim=False)

        assert type(arr) == aa.Array2D
        assert (arr == np.array([[0.0, 0.0], [0.0, 0.0]])).all()
        assert (arr.native == np.array([[0.0, 0.0], [0.0, 0.0]])).all()
        assert (arr.slim == np.array([0.0, 0.0, 0.0, 0.0])).all()

    def test__ones_zeros__makes_scaled_array_with_pixel_scale(self):

        arr = aa.Array2D.ones(shape_native=(2, 2), pixel_scales=1.0)

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 1.0], [1.0, 1.0]])).all()
        assert (arr.slim == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 0.0)

        arr = aa.Array2D.zeros(shape_native=(2, 2), pixel_scales=1.0, origin=(0.0, 1.0))

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[0.0, 0.0], [0.0, 0.0]])).all()
        assert (arr.slim == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

    def test__ones_zeros__makes_scaled_sub_array_with_pixel_scale_and_sub_size(self):

        arr = aa.Array2D.ones(shape_native=(1, 4), pixel_scales=1.0, sub_size=1)

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[1.0, 1.0, 1.0, 1.0]])).all()
        assert (arr.slim == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 0.0)
        assert arr.mask.sub_size == 1

        arr = aa.Array2D.zeros(
            shape_native=(1, 1), pixel_scales=1.0, sub_size=2, origin=(0.0, 1.0)
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.array([[0.0, 0.0], [0.0, 0.0]])).all()
        assert (arr.slim == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)
        assert arr.mask.sub_size == 2

    def test__from_fits__makes_array_without_other_inputs(self):

        arr = aa.Array2D.from_fits(
            file_path=path.join(test_data_dir, "3x3_ones.fits"), hdu=0, pixel_scales=1.0
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.ones((3, 3))).all()
        assert (arr.slim == np.ones(9)).all()

        arr = aa.Array2D.from_fits(
            file_path=path.join(test_data_dir, "4x3_ones.fits"),
            hdu=0,
            pixel_scales=1.0,
            store_slim=True,
        )

        assert type(arr) == aa.Array2D
        assert (arr == np.ones((12,))).all()
        assert (arr.native == np.ones((4, 3))).all()
        assert (arr.slim == np.ones((12,))).all()

        arr = aa.Array2D.from_fits(
            file_path=path.join(test_data_dir, "4x3_ones.fits"),
            hdu=0,
            pixel_scales=1.0,
            store_slim=False,
        )

        assert type(arr) == aa.Array2D
        assert (arr == np.ones((4, 3))).all()
        assert (arr.native == np.ones((4, 3))).all()
        assert (arr.slim == np.ones((12,))).all()

    def test__from_fits__makes_scaled_array_with_pixel_scale(self):

        arr = aa.Array2D.from_fits(
            file_path=path.join(test_data_dir, "3x3_ones.fits"), hdu=0, pixel_scales=1.0
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.ones((3, 3))).all()
        assert (arr.slim == np.ones(9)).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 0.0)

        arr = aa.Array2D.from_fits(
            file_path=path.join(test_data_dir, "4x3_ones.fits"),
            hdu=0,
            pixel_scales=1.0,
            origin=(0.0, 1.0),
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.ones((4, 3))).all()
        assert (arr.slim == np.ones((12,))).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

    def test__from_fits__makes_scaled_sub_array_with_pixel_scale_and_sub_size(self):

        arr = aa.Array2D.from_fits(
            file_path=path.join(test_data_dir, "3x3_ones.fits"),
            hdu=0,
            pixel_scales=1.0,
            sub_size=1,
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.ones((3, 3))).all()
        assert (arr.slim == np.ones(9)).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 0.0)
        assert arr.mask.sub_size == 1

        arr = aa.Array2D.from_fits(
            file_path=path.join(test_data_dir, "4x3_ones.fits"),
            hdu=0,
            pixel_scales=1.0,
            sub_size=1,
            origin=(0.0, 1.0),
        )

        assert type(arr) == aa.Array2D
        assert (arr.native == np.ones((4, 3))).all()
        assert (arr.slim == np.ones(12)).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)
        assert arr.mask.sub_size == 1

    def test__from_yx_values__use_manual_array_values__returns_input_array(self):

        arr = aa.Array2D.manual_native(array=[[1.0, 2.0], [3.0, 4.0]], pixel_scales=1.0)

        y = arr.mask.unmasked_grid_sub_1[:, 0]
        x = arr.mask.unmasked_grid_sub_1[:, 1]
        arr_via_yx = aa.Array2D.manual_yx_and_values(
            y=y, x=x, values=arr, shape_native=arr.shape_native, pixel_scales=1.0
        )

        assert (arr == arr_via_yx).all()

        arr = aa.Array2D.manual_native(
            array=[[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], pixel_scales=1.0
        )

        y = arr.mask.unmasked_grid_sub_1[:, 0]
        x = arr.mask.unmasked_grid_sub_1[:, 1]

        arr_via_yx = aa.Array2D.manual_yx_and_values(
            y=y, x=x, values=arr, shape_native=arr.shape_native, pixel_scales=1.0
        )

        assert (arr == arr_via_yx).all()

        arr = aa.Array2D.manual_native(
            array=[[1.0, 2.0, 3.0], [3.0, 4.0, 6.0]], pixel_scales=1.0
        )

        y = arr.mask.unmasked_grid_sub_1[:, 0]
        x = arr.mask.unmasked_grid_sub_1[:, 1]

        arr_via_yx = aa.Array2D.manual_yx_and_values(
            y=y, x=x, values=arr, shape_native=arr.shape_native, pixel_scales=1.0
        )

        assert (arr == arr_via_yx).all()

    def test__from_yx_values__use_input_values_which_swap_values_from_top_left_notation(
        self,
    ):

        arr = aa.Array2D.manual_yx_and_values(
            y=[0.5, 0.5, -0.5, -0.5],
            x=[-0.5, 0.5, -0.5, 0.5],
            values=[1.0, 2.0, 3.0, 4.0],
            shape_native=(2, 2),
            pixel_scales=1.0,
        )

        assert (arr.native == np.array([[1.0, 2.0], [3.0, 4.0]])).all()

        arr = aa.Array2D.manual_yx_and_values(
            y=[-0.5, 0.5, 0.5, -0.5],
            x=[-0.5, 0.5, -0.5, 0.5],
            values=[1.0, 2.0, 3.0, 4.0],
            shape_native=(2, 2),
            pixel_scales=1.0,
        )

        assert (arr.native == np.array([[3.0, 2.0], [1.0, 4.0]])).all()

        arr = aa.Array2D.manual_yx_and_values(
            y=[-0.5, 0.5, 0.5, -0.5],
            x=[0.5, 0.5, -0.5, -0.5],
            values=[1.0, 2.0, 3.0, 4.0],
            shape_native=(2, 2),
            pixel_scales=1.0,
        )

        assert (arr.native == np.array([[4.0, 2.0], [1.0, 3.0]])).all()

        arr = aa.Array2D.manual_yx_and_values(
            y=[1.0, 1.0, 0.0, 0.0, -1.0, -1.0],
            x=[-0.5, 0.5, -0.5, 0.5, -0.5, 0.5],
            values=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            shape_native=(3, 2),
            pixel_scales=1.0,
        )

        assert (arr.native == np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])).all()

        arr = aa.Array2D.manual_yx_and_values(
            y=[0.0, 1.0, -1.0, 0.0, -1.0, 1.0],
            x=[-0.5, 0.5, 0.5, 0.5, -0.5, -0.5],
            values=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            shape_native=(3, 2),
            pixel_scales=1.0,
        )

        assert (arr.native == np.array([[3.0, 2.0], [6.0, 4.0], [5.0, 1.0]])).all()



import autoarray as aa
import numpy as np


class TestDataVectorFromData:
    def test__simple_blurred_mapping_matrix__correct_data_vector(self):

        blurred_mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        image = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
        noise_map = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        data_vector = aa.util.inversion.data_vector_via_blurred_mapping_matrix_from(
            blurred_mapping_matrix=blurred_mapping_matrix,
            image=image,
            noise_map=noise_map,
        )

        assert (data_vector == np.array([2.0, 3.0, 1.0])).all()

    def test__simple_blurred_mapping_matrix__change_image_values__correct_data_vector(
        self,
    ):

        blurred_mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        image = np.array([3.0, 1.0, 1.0, 10.0, 1.0, 1.0])
        noise_map = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        data_vector = aa.util.inversion.data_vector_via_blurred_mapping_matrix_from(
            blurred_mapping_matrix=blurred_mapping_matrix,
            image=image,
            noise_map=noise_map,
        )

        assert (data_vector == np.array([4.0, 14.0, 10.0])).all()

    def test__simple_blurred_mapping_matrix__change_noise_values__correct_data_vector(
        self,
    ):

        blurred_mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        image = np.array([4.0, 1.0, 1.0, 16.0, 1.0, 1.0])
        noise_map = np.array([2.0, 1.0, 1.0, 4.0, 1.0, 1.0])

        data_vector = aa.util.inversion.data_vector_via_blurred_mapping_matrix_from(
            blurred_mapping_matrix=blurred_mapping_matrix,
            image=image,
            noise_map=noise_map,
        )

        assert (data_vector == np.array([2.0, 3.0, 1.0])).all()

    def test__data_vector_via_transformer_mapping_matrix_method__same_as_blurred_method_using_real_imag_separate(
        self,
    ):

        mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        data_real = np.array([4.0, 1.0, 1.0, 16.0, 1.0, 1.0])
        noise_map_real = np.array([2.0, 1.0, 1.0, 4.0, 1.0, 1.0])

        data_vector_real_via_blurred = aa.util.inversion.data_vector_via_blurred_mapping_matrix_from(
            blurred_mapping_matrix=mapping_matrix,
            image=data_real,
            noise_map=noise_map_real,
        )

        data_imag = np.array([4.0, 1.0, 1.0, 16.0, 1.0, 1.0])
        noise_map_imag = np.array([2.0, 1.0, 1.0, 4.0, 1.0, 1.0])

        data_vector_imag_via_blurred = aa.util.inversion.data_vector_via_blurred_mapping_matrix_from(
            blurred_mapping_matrix=mapping_matrix,
            image=data_imag,
            noise_map=noise_map_imag,
        )

        data_vector_complex_via_blurred = (
            data_vector_real_via_blurred + data_vector_imag_via_blurred
        )

        transformed_mapping_matrix = np.array(
            [
                [1.0 + 1.0j, 1.0 + 1.0j, 0.0 + 0.0j],
                [1.0 + 1.0j, 0.0 + 0.0j, 0.0 + 0.0j],
                [0.0 + 0.0j, 1.0 + 1.0j, 0.0 + 0.0j],
                [0.0 + 0.0j, 1.0 + 1.0j, 1.0 + 1.0j],
                [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
                [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
            ]
        )

        data = np.array(
            [4.0 + 4.0j, 1.0 + 1.0j, 1.0 + 1.0j, 16.0 + 16.0j, 1.0 + 1.0j, 1.0 + 1.0j]
        )
        noise_map = np.array(
            [2.0 + 2.0j, 1.0 + 1.0j, 1.0 + 1.0j, 4.0 + 4.0j, 1.0 + 1.0j, 1.0 + 1.0j]
        )

        data_vector_via_transformed = aa.util.inversion.data_vector_via_transformed_mapping_matrix_from(
            transformed_mapping_matrix=transformed_mapping_matrix,
            visibilities=data,
            noise_map=noise_map,
        )

        assert (data_vector_complex_via_blurred == data_vector_via_transformed).all()


class TestCurvatureMatrixFromBlurred:
    def test__simple_blurred_mapping_matrix(self):

        blurred_mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        noise_map = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        curvature_matrix = aa.util.inversion.curvature_matrix_via_mapping_matrix_from(
            mapping_matrix=blurred_mapping_matrix, noise_map=noise_map
        )

        assert (
            curvature_matrix
            == np.array([[2.0, 1.0, 0.0], [1.0, 3.0, 1.0], [0.0, 1.0, 1.0]])
        ).all()

    def test__simple_blurred_mapping_matrix__change_noise_values(self):

        blurred_mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        noise_map = np.array([2.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        curvature_matrix = aa.util.inversion.curvature_matrix_via_mapping_matrix_from(
            mapping_matrix=blurred_mapping_matrix, noise_map=noise_map
        )

        assert (
            curvature_matrix
            == np.array([[1.25, 0.25, 0.0], [0.25, 2.25, 1.0], [0.0, 1.0, 1.0]])
        ).all()


class TestPixelizationResiduals:
    def test__pixelization_perfectly_reconstructed_data__quantities_like_residuals_all_zeros(
        self,
    ):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.ones(9)
        slim_index_for_sub_slim_index = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        all_sub_slim_indexes_for_pixelization_index = [[0, 0, 0], [1, 1, 1], [2, 2, 2]]

        pixelization_residuals = aa.util.inversion.inversion_residual_map_from(
            pixelization_values=pixelization_values,
            data=reconstructed_data_1d,
            slim_index_for_sub_slim_index=slim_index_for_sub_slim_index,
            all_sub_slim_indexes_for_pixelization_index=all_sub_slim_indexes_for_pixelization_index,
        )

        assert (pixelization_residuals == np.zeros(3)).all()

    def test__pixelization_not_perfect_fit__quantities_like_residuals_non_zero(self):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = 2.0 * np.ones(9)
        slim_index_for_sub_slim_index = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        all_sub_slim_indexes_for_pixelization_index = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_residuals = aa.util.inversion.inversion_residual_map_from(
            pixelization_values=pixelization_values,
            data=reconstructed_data_1d,
            slim_index_for_sub_slim_index=slim_index_for_sub_slim_index,
            all_sub_slim_indexes_for_pixelization_index=all_sub_slim_indexes_for_pixelization_index,
        )

        assert (pixelization_residuals == 1.0 * np.ones(3)).all()

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.array([1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])
        slim_index_for_sub_slim_index = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        all_sub_slim_indexes_for_pixelization_index = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_residuals = aa.util.inversion.inversion_residual_map_from(
            pixelization_values=pixelization_values,
            data=reconstructed_data_1d,
            slim_index_for_sub_slim_index=slim_index_for_sub_slim_index,
            all_sub_slim_indexes_for_pixelization_index=all_sub_slim_indexes_for_pixelization_index,
        )

        assert (pixelization_residuals == np.array([0.0, 1.0, 2.0])).all()


class TestPixelizationNormalizedResiduals:
    def test__pixelization_perfectly_reconstructed_data__quantities_like_residuals_all_zeros(
        self,
    ):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.ones(9)
        noise_map_1d = np.ones(9)
        slim_index_for_sub_slim_index = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        all_sub_slim_indexes_for_pixelization_index = [[0, 0, 0], [1, 1, 1], [2, 2, 2]]

        pixelization_normalized_residuals = aa.util.inversion.inversion_normalized_residual_map_from(
            pixelization_values=pixelization_values,
            data=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            slim_index_for_sub_slim_index=slim_index_for_sub_slim_index,
            all_sub_slim_indexes_for_pixelization_index=all_sub_slim_indexes_for_pixelization_index,
        )

        assert (pixelization_normalized_residuals == np.zeros(3)).all()

    def test__pixelization_not_perfect_fit__quantities_like_residuals_non_zero(self):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = 2.0 * np.ones(9)
        noise_map_1d = 2.0 * np.ones(9)
        slim_index_for_sub_slim_index = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        all_sub_slim_indexes_for_pixelization_index = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_normalized_residuals = aa.util.inversion.inversion_normalized_residual_map_from(
            pixelization_values=pixelization_values,
            data=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            slim_index_for_sub_slim_index=slim_index_for_sub_slim_index,
            all_sub_slim_indexes_for_pixelization_index=all_sub_slim_indexes_for_pixelization_index,
        )

        assert (pixelization_normalized_residuals == 0.5 * np.ones(3)).all()

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.array([1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])
        noise_map_1d = np.array([0.5, 0.5, 0.5, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0])
        slim_index_for_sub_slim_index = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        all_sub_slim_indexes_for_pixelization_index = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_normalized_residuals = aa.util.inversion.inversion_normalized_residual_map_from(
            pixelization_values=pixelization_values,
            data=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            slim_index_for_sub_slim_index=slim_index_for_sub_slim_index,
            all_sub_slim_indexes_for_pixelization_index=all_sub_slim_indexes_for_pixelization_index,
        )

        assert (pixelization_normalized_residuals == np.array([0.0, 1.0, 1.0])).all()


class TestPixelizationChiSquareds:
    def test__pixelization_perfectly_reconstructed_data__quantities_like_residuals_all_zeros(
        self,
    ):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.ones(9)
        noise_map_1d = np.ones(9)
        slim_index_for_sub_slim_index = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        all_sub_slim_indexes_for_pixelization_index = [[0, 0, 0], [1, 1, 1], [2, 2, 2]]

        pixelization_chi_squareds = aa.util.inversion.inversion_chi_squared_map_from(
            pixelization_values=pixelization_values,
            data=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            slim_index_for_sub_slim_index=slim_index_for_sub_slim_index,
            all_sub_slim_indexes_for_pixelization_index=all_sub_slim_indexes_for_pixelization_index,
        )

        assert (pixelization_chi_squareds == np.zeros(3)).all()

    def test__pixelization_not_perfect_fit__quantities_like_residuals_non_zero(self):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = 2.0 * np.ones(9)
        noise_map_1d = 2.0 * np.ones(9)
        slim_index_for_sub_slim_index = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        all_sub_slim_indexes_for_pixelization_index = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_chi_squareds = aa.util.inversion.inversion_chi_squared_map_from(
            pixelization_values=pixelization_values,
            data=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            slim_index_for_sub_slim_index=slim_index_for_sub_slim_index,
            all_sub_slim_indexes_for_pixelization_index=all_sub_slim_indexes_for_pixelization_index,
        )

        assert (pixelization_chi_squareds == 0.25 * np.ones(3)).all()

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.array([1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])
        noise_map_1d = np.array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 4.0, 4.0, 4.0])
        slim_index_for_sub_slim_index = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        all_sub_slim_indexes_for_pixelization_index = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_chi_squareds = aa.util.inversion.inversion_chi_squared_map_from(
            pixelization_values=pixelization_values,
            data=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            slim_index_for_sub_slim_index=slim_index_for_sub_slim_index,
            all_sub_slim_indexes_for_pixelization_index=all_sub_slim_indexes_for_pixelization_index,
        )

        assert (pixelization_chi_squareds == np.array([0.0, 4.0, 0.25])).all()


class TestPreconditionerMatrix:
    def test__simple_calculations(self):

        mapping_matrix = np.array(
            [
                [1.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
                [0.0, 0.0, 1.0],
            ]
        )

        preconditioner_matrix = aa.util.inversion.preconditioner_matrix_via_mapping_matrix_from(
            mapping_matrix=mapping_matrix,
            preconditioner_noise_normalization=1.0,
            regularization_matrix=np.zeros((3, 3)),
        )

        assert (
            preconditioner_matrix
            == np.array([[2.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 2.0]])
        ).all()

        preconditioner_matrix = aa.util.inversion.preconditioner_matrix_via_mapping_matrix_from(
            mapping_matrix=mapping_matrix,
            preconditioner_noise_normalization=2.0,
            regularization_matrix=np.zeros((3, 3)),
        )

        assert (
            preconditioner_matrix
            == np.array([[4.0, 0.0, 0.0], [0.0, 4.0, 0.0], [0.0, 0.0, 4.0]])
        ).all()

        regularization_matrix = np.array(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]
        )

        preconditioner_matrix = aa.util.inversion.preconditioner_matrix_via_mapping_matrix_from(
            mapping_matrix=mapping_matrix,
            preconditioner_noise_normalization=2.0,
            regularization_matrix=regularization_matrix,
        )

        assert (
            preconditioner_matrix
            == np.array([[5.0, 2.0, 3.0], [4.0, 9.0, 6.0], [7.0, 8.0, 13.0]])
        ).all()
